from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


class NelderMead:
    r"""
    Downhill-Simplex algorithm for minimization.

    This implementation is based on the publication:
    Nelder and Mead, The Computer Journal 7, 308-313, 1965 (NM 1965)

    :Halting criterion:
    The default stop criterion is the one used by NM 1965. In
    particular, the value

    .. math:: \sqrt{\sum (\bar{y}-y_i) / n}

    is calculated. If it falls below the limit defined by the
    attribute `nmCritLim`, the iteration stops.

    Parameters
    ----------
    abg : tuple of three floats, optional
        Values for alpha, beta, and gamma.
    isw : float, optional
        Initial step width for simplex.
    critlim : float, optional
        Critical limit for stopping criterion.

    Attributes
    ----------
    alpha, beta, gamma : float
        The reflection-, expansion-, and contraction-coefficients.
        The default values (after NM 1965) are 1.0, 0.5, and 2.0.
        These coefficients control the modification of the simplex.
    initialStepWidthFac : float
        This factor determines how the initial simplex is calculated.
        The first simplex point is the starting value, the others are
        constructed by adding a fraction defined by this factor to
        the starting value. The default is 0.05.
    nmCritLim : float
        Critical value for the NM 1965 stopping criterion. The
        default is 1e-8.

    """

    def __init__(self, abg=(1.0, 0.5, 2.0), isw=0.05, critlim=1e-8):
        # Best guesses after NM 1965
        self.alpha, self.beta, self.gamma = abg
        # Step-width factor for initial simplex (if not specified otherwise)
        self.initialStepWidthFac = isw
        # Critical limit for the NM 1965 stopping criterion
        self.nmCritLim = critlim
        # Stopping criterion
        self._stopCrit = self._stopNM1965

    def _initSimplex(self, m, initDelta):
        r"""
        Define the initial simplex.

        Parameters
        ----------
        m : Instance of OneDFit
            The model.
        initDelta : dictionary
            Maps parameter name to the initial step width
            for the simplex.
        """
        # The simplex: n+1 points in n-dimensional space
        self._simplex = np.zeros((self._n + 1, self._n))
        # The function values
        self._yi = np.zeros(self._n + 1)
        # Define first point (order of parameters guaranteed)
        self._simplex[0, ::] = np.array(m.pars.getFreeParams())
        self._yi[0] = m.miniFunc(self._simplex[0, ::])
        for i, p in enumerate(self._fpns):
            self._simplex[i + 1, ::] = self._simplex[0, ::].copy()
            if (initDelta is not None) and (p in initDelta):
                self._simplex[i + 1, i] += initDelta[p]
            else:
                if self._simplex[0, i] == 0.0:
                    # Construct simplex by adding initial step width
                    self._simplex[i + 1, i] = self._simplex[0,
                                                            i] + self.initialStepWidthFac
                else:
                    # Construct simplex by adding fraction of the value
                    self._simplex[i + 1, i] += self._simplex[0,
                                                             i] * self.initialStepWidthFac
            self._yi[i + 1] = m.miniFunc(self._simplex[i + 1, ::])

    def _step(self, m):
        r"""
        Proceed one step of the simplex algorithm.
        """
        # Indices with lowest and highest objective function value
        h = np.argmax(self._yi)
        l = np.argmin(self._yi)
        # Barycenter (excluding the worst point)
        pb = (np.sum(self._simplex, 0) - self._simplex[h, ::]) / self._n
        # New suggestion
        ps = (1. + self.alpha) * pb - self.alpha * self._simplex[h, ::]
        ys = m.miniFunc(ps)
        if ys < self._yi[l]:
            # In this case, calculate pss
            pss = self.gamma * ps + (1. - self.gamma) * pb
            yss = m.miniFunc(pss)
            if yss < self._yi[l]:
                # Replace by pss
                self._simplex[h, ::] = pss
                self._yi[h] = yss
                return
            # Replace h by ps
            self._simplex[h, ::] = ps
            self._yi[h] = ys
            return
        # We have not produced a new minimum, and we know
        # that ys >= y[l]
        if ys < self._yi[h]:
            self._simplex[h, ::] = ps
            self._yi[h] = ys
            return
        else:
            # There is a new maximum
            pssb = self.beta * self._simplex[h, ::] + (1.0 - self.beta) * pb
            yssb = m.miniFunc(pssb)
            if yssb > self._yi[h]:
                # Contract
                pl = self._simplex[l, ::].copy()
                for i in smo.range(self._n + 1):
                    if i == l:
                        continue
                    self._simplex[i, ::] = (self._simplex[i, ::] + pl) / 2.0
                    self._yi[i] = m.miniFunc(self._simplex[i, ::])
                return
            else:
                self._simplex[h, ::] = pssb
                self._yi[h] = yssb
                return
        raise (PE.PyAOrderError(
            "Nothing done in Nelder-Mead step. If you see this, there is a coding error."))

    def _stopNM1965(self):
        r"""
        Apply the NM 1965 stopping criterion.

        Returns
        -------
        Reached : boolean
            True, if the stopping criterion has been reached.
        """
        ym = np.mean(self._yi)
        return (np.sqrt(np.sum((self._yi - ym)**2)) / self._n) < self.nmCritLim

    def fit(self, m, ds, objf="chisqr", initDelta=None, maxIter=1e4, callback=None, nmCritLim=None):
        r"""
        Carry out the model fit.

        After the iteration, the `iterCount` attribute contains the
        number of iterations. The `maxIterReached` attribute flag is
        False, if the maximum number of iterations has not been reached
        and True otherwise. 

        Parameters
        ----------
        m : Instance of OneDFit
            The model to be fitted.
        ds : Instance of FufDS
            The data.
        objf : string
            The objective function to be used. Possible
            choices are "chisqr" (default), "sqrdiff", and
            "cash79".
        initDelta : dictionary, optional
            A dictionary mapping parameter names to the
            initial step width. This can be very useful, if
            the starting values are zero or very small. The
            here defined step will be added to the starting
            value to construct the simplex.
        maxIter : int, optional
            The maximum number of iterations. The default is
            10000.
        nmCritLim : float, optional
            Critical value for stopping criterion. The default is
            1e-8.
        callback : callable, optional
            If not None, "callback" will be called with the
            three parameters: number of iteration (int), current
            best parameter set (array), and current simplex (array).

        Returns
        -------
        Best-fit values : dictionary
            Maps parameter name to the best-fit value.
        """
        # Stopping criterion
        if not nmCritLim is None:
            self.nmCritLim = nmCritLim
        # Number of free parameters
        self._n = m.numberOfFreeParams()
        # Set objective function
        m.setObjectiveFunction(objf)
        # Assign data object
        m._fufDS = ds
        # Names of free parameters (order guaranteed)
        self._fpns = m.freeParamNames()
        # Initial simplex
        self._initSimplex(m, initDelta)
        # MaxIter flag
        self.maxIterReached = False

        self.iterCount = 0
        while (not self._stopCrit()) and (self.iterCount < maxIter):
            self.iterCount += 1
            self._step(m)
            if callback is not None:
                l = np.argmin(self._yi)
                callback(self.iterCount, self._simplex[l, ::], self._simplex)

        # Find the optimum parameter set
        l = np.argmin(self._yi)
        m.pars.setFreeParams(self._simplex[l, ::])
        # Evaluate model so that model attribute holds the best match
        m.evaluate(ds.x)

        if self.iterCount == maxIter:
            self.maxIterReached = True
            PE.warn(PE.PyAAlgorithmFailure("The maximum number of iterations has been reached.\n" +
                                           "The fit may be inappropriate.",
                                           where="NelderMead",
                                           solution=["Increase number of iterations.",
                                                     "Change starting values.",
                                                     "Change algorithm parameters (e.g., alpha, beta, gamma)."]))
        # Return a dictionary with the best-bit parameters
        return dict(zip(self._fpns, self._simplex[l, ::]))



import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE

def quadExtreme(x, y, mode="max", dp=(1,1), exInd=None, fullOutput=False, fullPoint=False):
  """
    Find the extreme (minimum or maximum) by means of a parabolic fit.
    
    This function searches for the maximum (or minimum) in the given
    ordinate values, fits a second-order polynomial to the
    surroundings of that point, and, finally, determines the
    thus approximated abscissa value of the extreme point. 
    
    Parameters
    ----------
    x : array
        Abscissa values.
    y : array
        Ordinate values.
    mode : string, {min, max}, optional
        Determines whether a minimum or a maximum is searched
        for. The default is a maximum.
    dp : tuple with two integers, optional
        Determines the width around the extreme point, which will
        be used in the fitting. The default is one point to the
        right and left, i.e., dp = (1,1).
    extInd : integer, optional
        If given, the function will assume that this is the
        index of the extreme point and not search for it.
    fullOutput : boolean, optional
        If True, the output will also cover the points used in
        the fitting and the resulting polynomial.
    fullPoint : boolean, optional
        If True, the return value `epos` will be a tuple holding
        the abscissa and ordinate values of the extreme point.
        The default is False.
    
    Returns
    -------
    epos : float or tuple
        Position of the extreme point. If `fullPoint` is True,
        a tuple with the abscissa and ordinate values of the
        extreme point.
    mi : int
        The index of the extreme point (maximum or minimum).
    xb : array, optional
        Only returned if `fullOutput` is True. The abscissa
        values used in the polynomial fit. Note the the
        x-value of the extreme point has been subtracted.
    yb : array, optional
        Only returned if `fullOutput` is True. The ordinate
        values used in the polynomial fit.
    p : numpy polynomial, optional
        Only returned if `fullOutput` is True. The best-fit
        polynomial. Note that the fit refers to the `xb` axis,
        where the x-value of the extreme point has been
        subtracted.
  """
  if exInd is None:
    if mode == "max":
      mi = np.argmax(y)
    elif mode == "min":
      mi = np.argmin(y)
    else:
      raise(PE.PyAValError("Unknown mode '" + str(mode) + "'.", \
                           solution="Choose 'min' or 'max'"))
  else:
    mi = exInd

  if ((mi-dp[0]) < 0) or ((mi+dp[0]+1 > len(x))):
    raise(PE.PyAValError("The requested fitting range around the extreme is not covered by the data.", \
                         solution=["Adapt 'dp'.", "Do you need to change the mode?"]))
  
  # Cut out the relevant range
  xb = x[mi-dp[0]:mi+dp[1]+1].copy()
  yb = y[mi-dp[0]:mi+dp[1]+1].copy()
  
  # Subtract the x-value of the extreme
  xm = x[mi]
  xb -= xm
  
  # Fit second-order polynomial
  p = np.polyfit(xb, yb, 2)
  
  # Check whether curvature agrees with mode
  signs = {"max":-1, "min":+1}
  if np.sign(p[0]) != signs[mode]:
    PE.warn(PE.PyAValError("The curvature of the fit does not match the requested mode of '" + str(mode) + "'.", \
                           solution="Did you choose the right mode (min/max)?"))
  
  # Location of the extreme
  epos = -p[1]/(2.0 * p[0])
  # Shift back the location
  epos += xm
  
  if fullPoint:
    epos = (epos, -p[1]**2/(4.0*p[0]) + p[2])
  
  if not fullOutput:
    return epos, mi
  else:
    return epos, mi, xb, yb, p
    
  


# -*- coding: utf-8 -*-
from __future__ import print_function
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyaC import ImportCheck
import sys

# Check Python version
_majV, _minV = sys.version_info[0:2]
if (_minV < 7) and (_majV == 2):
    PE.warn("funcFit needs 2.7.x or greater. See documentation (Prerequisites) for explanation.")

ic = ImportCheck(["numpy", "scipy", "matplotlib",
                  "matplotlib.pylab", "pyfits", "emcee", "progressbar"])

# Get out if numpy not present
if not ic.check["numpy"]:
    raise(PE.PyARequiredImport("Numpy cannot be imported.", solution="Install numpy (see http://numpy.scipy.org/, you probably should also install SciPy).",
                               addInfo="The numpy package provides array support for Python and is indispensable in many scientific applications."))

# Check whether fitting modules can be imported
_scoImport = ic.check["scipy"]
_emceeImport = ic.check["emcee"]
_mplImport = ic.check["matplotlib"]

from PyAstronomy.funcFit.utils import *
from .fbo2 import *
from .anaMCMCTraces2 import TraceAnalysis2

def __question(tf):
    if tf:
        return "yes"
    return "no"


def status():
    print("Status of funcFit2:")
    print("--------------------------")
    print("Is scipy.optimize available? ", __question(_scoImport))
    if _scoImport:
        print("  Version: ", ic.versions["scipy"])
    if not _scoImport:
        print("  The minimization algorithms from SciPy cannot be used.")
        print("  Install SciPy to make them available (use 'pip install scipy' or see http://www.scipy.org/).")
    print("Is emcee available? ", __question(_emceeImport))
    if _emceeImport:
        print("  Version: ", ic.versions["emcee"])
    if not _emceeImport:
        print("  No Markov-Chain Monte-Carlo sampling via emcee will be possible.")
        print("  Install emcee to enable it (use 'pip install emcee' or see http://dfm.io/emcee/current/).")




# Original Author: Travis Oliphant 2002
# Bug-fixes in 2006 by Tim Leslie

# The content of this file is based on scipy.optimize's anneal.py file,
# which is distributed under the BSD license.

from __future__ import print_function, division
import numpy
from numpy import asarray, tan, exp, ones, squeeze, sign, \
    all, log, sqrt, pi, shape, array, minimum, where
from numpy import random
import numpy as np
import collections

from PyAstronomy.pyaC import pyaErrors as PE


_double_min = numpy.finfo(float).min
_double_max = numpy.finfo(float).max


class base_schedule(object):
    def __init__(self):
        self.dwell = 20
        self.learn_rate = 0.5
        self.lower = -10
        self.upper = 10
        self.Ninit = 50
        self.accepted = 0
        self.tests = 0
        self.feval = 0
        self.k = 0
        self.T = None

    def init(self, **options):
        self.__dict__.update(options)
        self.lower = asarray(self.lower)
        self.lower = where(self.lower == numpy.NINF, -_double_max, self.lower)
        self.upper = asarray(self.upper)
        self.upper = where(self.upper == numpy.PINF, _double_max, self.upper)
        self.k = 0
        self.accepted = 0
        self.feval = 0
        self.tests = 0

    def getstart_temp(self, best_state):
        """ Find a matching starting temperature and starting parameters vector
        i.e. find x0 such that func(x0) = T0.

        Parameters
        ----------
        best_state : _state
            A _state object to store the function value and x0 found.

        Returns
        -------
        x0 : array
            The starting parameters vector.
        r"""

        assert (not self.dims is None)
        lrange = self.lower
        urange = self.upper
        fmax = _double_min
        fmin = _double_max
        for _ in range(self.Ninit):
            x0 = random.uniform(size=self.dims)*(urange-lrange) + lrange
            fval = self.func(x0, *self.args)
            self.feval += 1
            if fval > fmax:
                fmax = fval
            if fval < fmin:
                fmin = fval
                best_state.cost = fval
                best_state.x = array(x0)

        self.T0 = (fmax-fmin)*1.5
        return best_state.x

    def accept_test(self, dE):
        T = self.T
        self.tests += 1
        if dE < 0:
            self.accepted += 1
            return 1
        p = exp(-dE*1.0/self.boltzmann/T)
        if (p > random.uniform(0.0, 1.0)):
            self.accepted += 1
            return 1
        return 0

    def update_guess(self, x0):
        pass

    def update_temp(self, x0):
        pass


#  A schedule due to Lester Ingber
class fast_sa(base_schedule):

    def init(self, **options):
        self.__dict__.update(options)
        if self.m is None:
            self.m = 1.0
        if self.n is None:
            self.n = 1.0
        self.c = self.m * exp(-self.n * self.quench)

    def update_guess(self, x0):
        x0 = np.asarray(x0)
        u = np.squeeze(np.random.uniform(0.0, 1.0, size=self.dims))
        T = self.T
        y = np.sign(u-0.5)*T*((1+1.0/T)**np.abs(2*u-1)-1.0)
        xc = y*(self.upper - self.lower)
        xnew = x0 + xc
        return xnew

    def update_temp(self):
        self.T = self.T0*exp(-self.c * self.k**(self.quench))
        self.k += 1
        return


class cauchy_sa(base_schedule):
    def update_guess(self, x0):
        x0 = asarray(x0)
        numbers = squeeze(random.uniform(-pi/2, pi/2, size=self.dims))
        xc = self.learn_rate * self.T * tan(numbers)
        xnew = x0 + xc
        return xnew

    def update_temp(self):
        self.T = self.T0/(1+self.k)
        self.k += 1
        return


class boltzmann_sa(base_schedule):

    def update_guess(self, x0):
        """
        r"""
        std = minimum(sqrt(self.T)*ones(self.dims),
                      (self.upper-self.lower)/3.0/self.learn_rate)
        x0 = asarray(x0)
        xc = squeeze(random.normal(0, 1.0, size=self.dims))

        xnew = x0 + xc*std*self.learn_rate
        return xnew

    def update_temp(self):
        """
        Update the temperature using the Boltzmann schedule
        r"""
        self.k += 1
        self.T = self.T0 / log(self.k+1.0)


class PyASASBase:

    def __init__(self, fo, T0, Tf, dwell, pdscale=None):
        self._fo = fo
        # Assign starting temperature, "running" temperature,
        # and final temperature
        self.T0 = T0
        self.T = self.T0
        self.Tf = Tf
        # Count the temperature updates
        self.tempUpdates = 0
        # Assign the dwell parameter
        self.dwell = dwell
        # Assign the proposal density scaling
        self.pdscale = {}
        for p in fo.freeParamNames():
            self.pdscale[p] = 1.0
            if pdscale is not None:
                if p in pdscale:
                    self.pdscale[p] = pdscale[p]
        # Count the number of parameter updates
        self.guessCounter = 0
        # Count accepted and rejected moves
        self.accepted = 0
        self.rejected = 0

    def updateGuess(self):
        raise (PE.PyANotImplemented("The 'updateGuess' is not implemented.",
                                    solution="Use another scheduler class (e.g., 'BoltzmannSAS') or implement the method."))

    def finalTemperatureReached(self):
        """
        Check whether the final temperature has been reached.

        Returns
        -------
        Reached flag : boolean
          True if the final temperature has been reached.
        r"""
        # If no final temperature has been specified,
        # it cannot be reached.
        if self.Tf is None:
            return False
        # Else check whether it is the case
        if self.T < self.Tf:
            return True
        return False

    def updateTemperature(self):
        raise (PE.PyANotImplemented("The 'updateTemperature' is not implemented.",
                                    solution="Use another scheduler class (e.g., 'BoltzmannSAS') or implement the method."))

    def acceptanceTest(self, dE):
        raise (PE.PyANotImplemented("The 'acceptanceTest' is not implemented.",
                                    solution="Use another scheduler class (e.g., 'BoltzmannSAS') or implement the method."))


class BoltzmannSAS(PyASASBase):
    """
    Scheduler for Boltzmann annealing.

    Parameters
    ----------
    fo : OneDFit object
        The fitting object for which to propose a new
        parameter set.
    pdscale : dictionary, optional
        Scale for the proposal density for the individual
        parameters.
    r"""

    def __init__(self, fo, T0, pdscale=None, dwell=20, Tf=None):
        PyASASBase.__init__(self, fo, T0, Tf, dwell, pdscale)

    def updateGuess(self):
        """
        Propose a new position in state space.

        Returns
        -------
        New parameters : dictionary
            A dictionary relating parameter name and new
            value.
        r"""
        x0 = self._fo.freeParameters()
        y0 = x0.copy()
        for p in x0.keys():
            y0[p] += (np.random.normal(0., 1.) *
                      self.pdscale[p] * np.sqrt(self.T/self.T0))
        self.guessCounter += 1
        return y0

    def updateTemperature(self):
        """
        Recalculate the annealing temperature.
        r"""
        n = self.accepted + self.rejected
        if (n % self.dwell == 0) and (n != 0):
            # Update only after the dwell period
            self.tempUpdates += 1
            self.T = self.T0 / np.log(self.tempUpdates + 1)

    def acceptanceTest(self, dE):
        """
        Test whether a step shall be accepted or rejected.

        Parameters
        ----------
        dE : float
            The "energy difference", i.e., the difference
            in the cost function.

        Returns
        -------
        Acceptance flag : boolean
            True if the step is accepted.
        r"""
        if dE < 0:
            # If the proposed state is better, accept
            # it in any case
            self.accepted += 1
            return True
        # Accept the step with temperature-dependent
        # probability
        p = 1.0 / (1.0 + np.exp(dE/self.T))
        if (p > random.uniform(0.0, 1.0)):
            self.accepted += 1
            return True
        self.rejected += 1
        return False


class CauchySAS(PyASASBase):
    """
    r"""

    def __init__(self, fo, T0, pdscale=None, dwell=20, Tf=None):
        PyASASBase.__init__(self, fo, T0, Tf, dwell, pdscale)


class _state(object):
    def __init__(self):
        self.x = None
        self.cost = None

# TODO:
#     allow for general annealing temperature profile
#     in that case use update given by alpha and omega and
#     variation of all previous updates and temperature?


class PyAAnneal:

    def __init__(self, fo):
        self._fo = fo
        self.scheduler = None
        self.states = {"current": {}, "last": {}, "current": {}}
        self.functionEvaluations = 0
        # Determines whether a stopping criterion had been reached
        # and which one.
        self.stopEncountered = False
        self.whichStop = "No stop"
        # The maximum number of iterations
        self.maxiter = 100000
        # Info parameters
        self.statsPeriod = 1000
        self.verbose = True

    def anneal(self, miniFunc, scheduler):
        """
        r"""

        self.scheduler = scheduler
        print("miniFunc: ", miniFunc)
        self.states["last"]["pars"] = self._fo.freeParameters()
        self.states["last"]["cost"] = miniFunc(self._fo.pars.getFreeParams())
        # At this moment, the best state IS the last state
        self.states["best"] = self.states["last"].copy()

        # Initialize iteration counter
        iterations = 0

        # Reset stopping information
        self.stopEncountered = False
        self.whichStop = "No stop"

        # Save the last n cost values to check whether
        # there is still improvement
        costDeque = collections.deque([], self.statsPeriod)
        # Variables to calculate acceptance rate
        lastAccept = 0
        lastReject = 0

        while True:
            # Save parameters
            savePars = self._fo.parameters()
            # Get new parameter guess and assign
            self.states["current"]["pars"] = self.scheduler.updateGuess()
            self._fo.assignValues(self.states["current"]["pars"])
            # Evaluate objective function
            self.states["current"]["cost"] = miniFunc(
                self._fo.pars.getFreeParams())

            dE = self.states["current"]["cost"] - self.states["last"]["cost"]
            if self.scheduler.acceptanceTest(dE):
                # The step has been accepted
                self.states["last"] = self.states["current"].copy()
                if self.states["last"]["cost"] < self.states["best"]["cost"]:
                    self.states["best"] = self.states["last"].copy()
            else:
                # Reset parameter values in case of rejected step
                self._fo.assignValues(savePars)

            self.scheduler.updateTemperature()
            iterations += 1

            if self.scheduler.finalTemperatureReached():
                self.stopEncountered = True
                self.whichStop = "Final temperature of " + \
                    str(self.scheduler.Tf) + " has been reached."
                return
            if iterations > self.maxiter:
                self.stopEncountered = True
                self.whichStop = "Maximum number of iterations (" + str(
                    self.maxiter) + ") was reached."
                return

            costDeque.append(self.states["current"]["cost"])

            if self.verbose:
                if (iterations % self.statsPeriod == 0) and (iterations != 0):
                    print("After " + str(iterations) + " iterations:")
                    print("  Best solution: ")
                    print("    Parameters: ", self.states["best"]["pars"])
                    print("    Objective function value: ",
                          self.states["best"]["cost"])
                    print("  Current temperature: ", self.scheduler.T)
                    print("  Average objective function in last period: ",
                          np.mean(costDeque))
                    print("  Acceptance rate in last period: ",
                          float(self.scheduler.accepted-lastAccept) /
                          (self.scheduler.accepted + self.scheduler.rejected - lastAccept - lastReject))
                    lastAccept = self.scheduler.accepted
                    lastReject = self.scheduler.rejected


# Simulated annealing

def anneal(func, x0, args=(), schedule='fast', full_output=0,
           T0=None, Tf=1e-12, maxeval=None, maxaccept=None, maxiter=400,
           feps=1e-6, quench=1.0, m=1.0, n=1.0,
           dwell=50):
    """Minimize a function using simulated annealing.

    Schedule is a schedule class implementing the annealing schedule.
    Available ones are 'fast', 'cauchy', 'boltzmann'

    Parameters
    ----------
    func : callable f(x, *args)
        Function to be optimized.
    x0 : ndarray
        Initial guess.
    args : tuple
        Extra parameters to `func`.
    schedule : base_schedule
        Annealing schedule to use (a class).
    full_output : bool
        Whether to return optional outputs.
    T0 : float
        Initial Temperature (estimated as 1.2 times the largest
        cost-function deviation over random points in the range).
    Tf : float
        Final goal temperature.
    maxeval : int
        Maximum function evaluations.
    maxaccept : int
        Maximum changes to accept.
    maxiter : int
        Maximum cooling iterations.
    feps : float
        Stopping relative error tolerance for the function value in
        last four coolings.
    quench, m, n : float
        Parameters to alter fast_sa schedule.
    dwell : int
        The number of times to search the space at each temperature.

    Outputs
    -------
    xmin : ndarray
        Point giving smallest value found.
    retval : int
        Flag indicating stopping condition::

                0 : Cooled to global optimum
                1 : Cooled to final temperature
                2 : Maximum function evaluations
                3 : Maximum cooling iterations reached
                4 : Maximum accepted query locations reached

    Jmin : float
        Minimum value of function found.
    T : float
        Final temperature.
    feval : int
        Number of function evaluations.
    iters : int
        Number of cooling iterations.
    accept : int
        Number of tests accepted.

    """
    x0 = asarray(x0)

    schedule = eval(schedule+'_sa()')
    #   initialize the schedule
    schedule.init(dims=shape(x0), func=func, args=args, T0=T0,
                  m=m, n=n, quench=quench, dwell=dwell)

    current_state, last_state, best_state = _state(), _state(), _state()
    if T0 is None:
        x0 = schedule.getstart_temp(best_state)
    else:
        best_state.x = None
        best_state.cost = 300e8

    last_state.x = asarray(x0).copy()
    fval = func(x0, *args)
    schedule.feval += 1
    last_state.cost = fval
    if last_state.cost < best_state.cost:
        best_state.cost = fval
        best_state.x = asarray(x0).copy()
    schedule.T = schedule.T0
    fqueue = [100, 300, 500, 700]
    iters = 0
    while 1:
        for n in range(dwell):
            current_state.x = schedule.update_guess(last_state.x)
            current_state.cost = func(current_state.x, *args)
            schedule.feval += 1

            dE = current_state.cost - last_state.cost
            if schedule.accept_test(dE):
                last_state.x = current_state.x.copy()
                last_state.cost = current_state.cost
                if last_state.cost < best_state.cost:
                    best_state.x = last_state.x.copy()
                    best_state.cost = last_state.cost
        schedule.update_temp()
        iters += 1
        # Stopping conditions
        # 0) last saved values of f from each cooling step
        #     are all very similar (effectively cooled)
        # 1) Tf is set and we are below it
        # 2) maxeval is set and we are past it
        # 3) maxiter is set and we are past it
        # 4) maxaccept is set and we are past it

        fqueue.append(squeeze(last_state.cost))
        fqueue.pop(0)
        af = asarray(fqueue)*1.0
        if all(abs((af-af[0])/af[0]) < feps):
            retval = 0
            if abs(af[-1]-best_state.cost) > feps*10:
                retval = 5
                print("Warning: Cooled to %f at %s but this is not"
                      % (squeeze(last_state.cost), str(squeeze(last_state.x)))
                      + " the smallest point found.")
            break
        if (Tf is not None) and (schedule.T < Tf):
            retval = 1
            break
        if (maxeval is not None) and (schedule.feval > maxeval):
            retval = 2
            break
        if (iters > maxiter):
            print("Warning: Maximum number of iterations exceeded.")
            retval = 3
            break
        if (maxaccept is not None) and (schedule.accepted > maxaccept):
            retval = 4
            break

    if full_output:
        return best_state.x, best_state.cost, schedule.T, \
            schedule.feval, iters, schedule.accepted, retval
    else:
        return best_state.x, retval


if __name__ == "__main__":
    from numpy import cos
    # minimum expected at ~-0.195
    def func(x): return cos(14.5*x-0.3) + (x+0.2)*x
    print(anneal(func, 1.0, full_output=1, upper=3.0, lower=-
          3.0, feps=1e-4, maxiter=2000, schedule='cauchy'))
    print(anneal(func, 1.0, full_output=1, upper=3.0, lower=-
          3.0, feps=1e-4, maxiter=2000, schedule='fast'))
    print(anneal(func, 1.0, full_output=1, upper=3.0, lower=-
          3.0, feps=1e-4, maxiter=2000, schedule='boltzmann'))

    # minimum expected at ~[-0.195, -0.1]
    def func(x): return cos(14.5*x[0]-0.3) + (x[1]+0.2)*x[1] + (x[0]+0.2)*x[0]
    print(anneal(func, [1.0, 1.0], full_output=1, upper=[
          3.0, 3.0], lower=[-3.0, -3.0], feps=1e-4, maxiter=2000, schedule='cauchy'))
    print(anneal(func, [1.0, 1.0], full_output=1, upper=[
          3.0, 3.0], lower=[-3.0, -3.0], feps=1e-4, maxiter=2000, schedule='fast'))
    print(anneal(func, [1.0, 1.0], full_output=1, upper=[
          3.0, 3.0], lower=[-3.0, -3.0], feps=1e-4, maxiter=2000, schedule='boltzmann'))



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy import funcFit as fuf
import scipy.interpolate as sci
from quantities.dimensionality import assert_isinstance


def equidistantInterpolation(x, y, dxmode, ifct=sci.interp1d):
    """
    Construct evenly sampled data set by interpolation
    
    Parameters
    ----------
    x : array
        Input x-axis
    y : array or list/tuple of arrays
        The y axis or axes
    dxmode : string {2x, mean} or float
        If a float is given, it specifies the spacing of the new x axis. If '2x' is
        given, an equidistant x axis with twice as many points as the input one will
        be used. If 'mean' is given, the spacing of the new x axis corresponds to the
        mean of that of the input axis. If dxmode is an array, it is used as the new x-axis
        (without check for equidistant sampling).
    ifct : interpolation callable
        A callable taking x and y as arguments and returning a callable, which takes
        the new equidistant wavelength axis as argument. The default is interp1d from
        scipy called by ny = ifct(x, y)(nx).
    
    Returns
    -------
    x, y : arrays
        New equidistant x axis and linearly interpolated y axis. If input y is a list or tuple of arrays,
        a list of interpolated arrays is returned.
    """
    if type(dxmode) == float:
        nx = np.arange(x[0], x[-1], dxmode)
    elif isinstance(dxmode, np.ndarray):
        nx = dxmode
    elif dxmode == "2x":
        nx = np.linspace(x[0], x[-1], 2*len(x))
    elif dxmode == "mean":
        dx = np.mean(np.diff(x))
        nx = np.arange(x[0], x[-1], dx)
    else:
        raise(PE.PyAValError("Cannot interpret 'dxmode'", \
                             where="makeEquidistant", \
                             solution=["Use a float to specify spacing.", \
                                       "Use '2x' or 'mean' to give rule to select spacing"]))
    
    if isinstance(y, (list,tuple)):
        ny = [ifct(x, yy)(nx) for yy in y]
    else:
        ny = ifct(x, y)(nx)
    
    return nx, ny


def broadGaussFast(x, y, sigma, edgeHandling=None, maxsig=None):
    """
    Apply Gaussian broadening. 

    This function broadens the given data using a Gaussian
    kernel.

    Parameters
    ----------
    x, y : arrays
        The abscissa and ordinate of the data.
    sigma : float
        The width (i.e., standard deviation) of the Gaussian
        profile used in the convolution.
    edgeHandling : string, {None, "firstlast"}, optional
        Determines the way edges will be handled. If None,
        nothing will be done about it. If set to "firstlast",
        the spectrum will be extended by using the first and
        last value at the start or end. Note that this is
        not necessarily appropriate. The default is None.
    maxsig : float, optional
        The extent of the broadening kernel in terms of
        standard deviations. By default, the Gaussian broadening
        kernel will be extended over the entire given spectrum,
        which can cause slow evaluation in the case of large spectra.
        A reasonable choice could, e.g., be five.  

    Returns
    -------
    Broadened data : array
        The input data convolved with the Gaussian
        kernel.
    """
    # Check whether x-axis is linear
    dxs = x[1:] - x[0:-1]
    if abs(max(dxs) - min(dxs)) > np.mean(dxs) * 1e-6:
        raise(PE.PyAValError("The x-axis is not equidistant, which is required.",
                             where="broadGaussFast",
                             solution=["Use pyasl.equidistantInterpolation to produce equidistantly sampled data."]))

    if maxsig is None:
        lx = len(x)
    else:
        lx = int(((sigma * maxsig) / dxs[0]) * 2.0) + 1
    # To preserve the position of spectral lines, the broadening function
    # must be centered at N//2 - (1-N%2) = N//2 + N%2 - 1
    nx = (np.arange(lx, dtype=int) - sum(divmod(lx, 2)) + 1) * dxs[0]
    gf = fuf.GaussFit1d()
    gf["A"] = 1.0
    gf["sig"] = sigma
    e = gf.evaluate(nx)
    # This step ensured that the
    e /= np.sum(e)

    if edgeHandling == "firstlast":
        nf = len(y)
        y = np.concatenate((np.ones(nf) * y[0], y, np.ones(nf) * y[-1]))
        result = np.convolve(y, e, mode="same")[nf:-nf]
    elif edgeHandling is None:
        result = np.convolve(y, e, mode="same")
    else:
        raise(PE.PyAValError("Invalid value for `edgeHandling`: " + str(edgeHandling),
                             where="broadGaussFast",
                             solution="Choose either 'firstlast' or None"))
    return result


def instrBroadGaussFast(wvl, flux, resolution, edgeHandling=None, fullout=False, maxsig=None, equid=False):
    """
    Apply Gaussian instrumental broadening. 

    This function broadens a spectrum assuming a Gaussian
    kernel. The width of the kernel is determined by the
    resolution. In particular, the function will determine
    the mean wavelength and set the Full Width at Half
    Maximum (FWHM) of the Gaussian to
    (mean wavelength)/resolution. 

    Parameters
    ----------
    wvl : array
        The wavelength
    flux : array
        The spectrum
    resolution : int
        The spectral resolution.
    edgeHandling : string, {None, "firstlast"}, optional
        Determines the way edges will be handled. If None,
        nothing will be done about it. If set to "firstlast",
        the spectrum will be extended by using the first and
        last value at the start or end. Note that this is
        not necessarily appropriate. The default is None.
    fullout : boolean, optional
        If True, also the FWHM of the Gaussian will be returned.
    maxsig : float, optional
        The extent of the broadening kernel in terms of
        standard deviations. By default, the Gaussian broadening
        kernel will be extended over the entire given spectrum,
        which can cause slow evaluation in the case of large spectra.
        A reasonable choice could, e.g., be five.
    equid : boolean, optional
        If True, linear interpolation will be used to obtain an
        intermediate spectrum with uniformly sampled wavelength grid
        to carry out the broadening.

    Returns
    -------
    Broadened spectrum : array
        The input spectrum convolved with a Gaussian
        kernel.
    FWHM : float, optional
        The Full Width at Half Maximum (FWHM) of the
        used Gaussian kernel.
    """
    # Check whether wvl axis is linear
    dwls = wvl[1:] - wvl[0:-1]
    # True if wavelength axis is not equidistant
    dwlned = abs(max(dwls) - min(dwls)) > np.mean(dwls) * 1e-6
    if dwlned and (not equid):
        raise(PE.PyAValError("The wavelength axis is not equidistant, which is required.",
                             where="instrBroadGaussFast", \
                             solution=["Define equidistant wavelength axis", "Set 'equid' to True",
                                       "Use pyasl.equidistantInterpolation to produce equidistantly sampled data."]))
    elif dwlned and equid:
        # Define an equidistant wavelength grid
        wvlorig = wvl
        wvl = np.linspace(wvl[0], wvl[-1], 2*len(wvl))
        flux = sci.interp1d(wvlorig, flux)(wvl)
    elif (not dwlned) and equid:
        # Wavelength axis is equidistant and equid is also True (not required) 
        equid = False
    elif (not dwlned) and (not equid):
        # Wavelength axis is equidistant and no intermediate wvl axis requested
        pass
        
    meanWvl = np.mean(wvl)
    fwhm = 1.0 / float(resolution) * meanWvl
    sigma = fwhm / (2.0 * np.sqrt(2. * np.log(2.)))

    result = broadGaussFast(
        wvl, flux, sigma, edgeHandling=edgeHandling, maxsig=maxsig)

    if equid:
        # Recover wvl grid
        result = sci.interp1d(wvl, result)(wvlorig)

    if not fullout:
        return result
    else:
        return (result, fwhm)


def thermalBroadeningWidth(lam0, T, m=None, fwhm=True):
    """
    Calculate the width for thermal broadening.

    Thermal motion of particles causes a Doppler
    broadening of the line profile. The resulting
    line profile is Gaussian with FWHM given by

    .. math::

        fwhm = \\lambda_0\\sqrt{\\frac{8k_B T\\ln(2)}{m c^2}}

    See, e.g.,
    http://hyperphysics.phy-astr.gsu.edu/hbase/atomic/broaden.html

    Parameters
    ----------
    lam0 : float
        Wavelength at which to calculate the width.
    T : float
        Temperature [K].
    m : float, optional
        Mass of the particles [kg]. If not specified,
        the proton mass is assumed.
    fwhm : boolean, optional
        If True (default), the FWHM of the Gaussian
        broadening kernel will be returned. Otherwise,
        the standard deviation is returned.

    Returns
    -------
    Width : float
        The width of the Gaussian broadening kernel.
        By default, the FWHM is returned. To obtain
        the standard deviation, set the `fwhm` flag to
        False.
    """
    from PyAstronomy import constants as PC
    pc = PC.PyAConstants()
    pc.setSystem("SI")
    if m is None:
        # Use proton mass if not specified otherwise
        m = pc.mp
    result = lam0 * np.sqrt(8.0 * pc.k * T * np.log(2.0) / (m * (pc.c)**2))
    if not fwhm:
        return result / (2.0 * np.sqrt(2.0 * np.log(2.0)))
    else:
        return result
    
    
    
def tempFromthermalBroadeningWidth(lam0, width, m=None, awidth=0.0):
    """
    Calculate the temperature required to obtain thermal broadening width.

    Thermal motion of particles causes Doppler
    broadening of the line profile. The resulting
    line profile is Gaussian with FWHM given by

    .. math::

        fwhm = \\lambda_0\\sqrt{\\frac{8k_B T\\ln(2)}{m c^2}}

    See, e.g.,
    http://hyperphysics.phy-astr.gsu.edu/hbase/atomic/broaden.html

    Here, the relation is reversed, so that the temperature, T, can be
    obtained from the width. The relation is only strictly valid in the case
    of Gaussian lines.

    .. note:: Units of lam0, width, and awidth have to be consistent.

    Parameters
    ----------
    lam0 : float
        Wavelength at which to width (FWHM) was measured.
    width : float
        Measured width of the feature (FWHM) 
    m : float, optional
        Mass of the particles [kg]. If not specified,
        the proton mass is assumed.
    awidth : float
        Additional width contributed by other effects such as
        instrumental resolution. The thermal width, ft, is obtained by
        quadratic subtraction, i.e., ft**2 = width**2 - awidth**2.
        Default is zero.

    Returns
    -------
    Temperature : float
        The temperature required to achieve the specified broadening
        width.
    Thermal width : float
        The thermal width used in the calculations (may be modified
        if `awidth` has been specified.
    """
    from PyAstronomy import constants as PC
    pc = PC.PyAConstants()
    pc.setSystem("SI")
    if m is None:
        # Use proton mass if not specified otherwise
        m = pc.mp
        
    if awidth > width:
        raise(PE.PyAValError("The additional width (awidth) is larger than the specified width of the feature.", \
                             where="tempFromthermalBroadeningWidth", \
                             solution="Check width specification.")) 
        
    # Take into account additional width
    width = np.sqrt(width**2 - awidth**2)
        
    result = width**2 / lam0**2 * (m * (pc.c)**2) / (8.0 * pc.k * np.log(2.0))
    return result
    




# -*- coding: utf-8 -*-
from __future__ import print_function, division
import numpy
from .onedfit import OneDFit
import scipy.special as sps
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


class Voigt1d(OneDFit):
    r"""
    Implements a Voigt profile (convolution of Cauchy-Lorentz
    and Gaussian distribution).

    .. note:: The profile is implemented so that `al` is half
              the FWHM of the Cauchy-Lorentz distribution.

    *Fit parameters*:
      - `A` - Area under the curve
      - `al` - Scale parameter of the Cauchy-Lorentz distribution (half its FWHM)
      - `ad` - The width (standard deviation) of the Gaussian (usually called sigma)
      - `mu` - Center
      - `off` - Constant offset
      - `lin` - Linear contribution

    Notes
    -----

    The Voigt profile V is defined as the convolution

    .. math::
        V(x) = A\int G(x')L(x-x')dx'

    of a Gaussian distribution

    .. math::
        G=1/(\sqrt{2 \pi} \ ad)\exp(-(x-mu)^2/(2 \ ad^2))

    and a Cauchy-Lorentz distribution

    .. math::
        L=al/(\pi ((x-mu)^2+al^2)) .

    We here take into account an additional offset and linear term so that

    .. math::
        V'(x) = V(x) + (lin \ x + off) .

    The Voigt profile is calculated via the real part of
    the Faddeeva function.
    For details, see http://en.wikipedia.org/wiki/Voigt_profile 
    and http://en.wikipedia.org/wiki/Error_function.
    """

    def __init__(self):
        OneDFit.__init__(self, ["A", "al", "ad", "mu", "lin", "off"])
        self.setRestriction({"al": [0.0, None], "ad": [0.0, None]})
        self.setRootName("Voigt")

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        try:

            z = ((x - self["mu"]) + (1.j) * abs(self["al"])) / \
                (numpy.sqrt(2.0) * abs(self["ad"]))
            y = self["A"] * numpy.real(sps.wofz(z))
            y /= (abs(self["ad"]) * numpy.sqrt(2.0 * numpy.pi))
            y += x * self["lin"] + self["off"]

            y[numpy.where(numpy.isnan(y))] = 0.0
        except FloatingPointError as fpe:
            raise (PE.PyAFloatingPointError("The following floating point error occurred:\n  " + str(fpe) + "\n" +
                                            "Current Parameter values:\n" +
                                            str(self.parameters()),
                                            solution=["Try to rescale/shift your abscissa. For instance, put" +
                                                      "the spectral line you try to fit at position `0`."]))
        return y

    def FWHM(self):
        r"""
        Calculates an approximation of the FWHM.

        The approximation is accurate to 
        about 0.03% (see http://en.wikipedia.org/wiki/Voigt_profile).

        Returns
        -------
        FWHM : float
            The estimate of the FWHM
        """
        # The width of the Lorentz profile
        fl = 2.0 * self["al"]
        # Width of the Gaussian [2.35 = 2*sigma*sqrt(2*ln(2))]
        fd = 2.35482 * self['ad']
        return 0.5346 * fl + numpy.sqrt(0.2166 * (fl**2.) + fd**2.)


class MultiVoigt1d(OneDFit):
    r"""
    Multicomponent Voigt with a single linear continuum component.

    The parameters are the same as for *Voigt1d*,
    except that all are extended by a number specifying the
    component to which they belong. Therefore, they read, e.g.,
    `A1`, `mu2`, and so on; only `off` and `lin`
    remain unnumbered.

    Parameters
    ----------
    n : int
        The number of Voigt components.
    """

    def __init__(self, n):
        # Number of components
        self.n = n
        # Building parameter list
        params = ["off", "lin"]
        for i in range(n):
            p = str(i + 1)
            params.extend(["A" + p, "mu" + p, "al" + p, "ad" + p])
        OneDFit.__init__(self, params)
        self.setRootName("MultiVoigt")
        # Use Voigt1d for evaluation
        self._v1d = Voigt1d()
        self._v1d["off"] = 0.0
        self._v1d["lin"] = 0.0

    def evaluate(self, x):
        r"""
        Evaluates the model for current parameter values.

        Parameters
        ----------
        x : array
            Specifies the points at which to evaluate the model.
        """
        # Assign 'continuum'
        y = self["off"] + self["lin"] * x
        # Add Voigt lines
        for i in smo.range(self.n):
            p = str(i + 1)
            self._v1d.assignValues(
                {"A": self["A" + p], "al": self["al" + p], "ad": self["ad" + p], "mu": self["mu" + p]})
            y += self._v1d.evaluate(x)
        return y

    def evalComponent(self, x, p):
        r"""
        Evaluate the model considering only a single component.

        Parameters
        ----------
        x : array
            The abscissa.
        p : int
            Component number (starts with one).

        Returns
        -------
        Single component model : array
            The model considering a single component. Note that the
            linear continuum is included.
        """
        if p > 0 and p <= self.n:
            p = str(p)
            y = self["off"] + self["lin"] * x
            self._v1d.assignValues(
                {"A": self["A" + p], "al": self["al" + p], "ad": self["ad" + p], "mu": self["mu" + p]})
            y += self._v1d.evaluate(x)
            return y
        else:
            raise (PE.PyAValError("No such component (no. " + str(p) + ")", where="MultiVoigt1d::evalComponent",
                                  solution="Use value between 1 and " + str(self.n)))



from __future__ import print_function, division
import numpy as np
import os
from PyAstronomy.pyaC import pyaErrors as PE
import six.moves as smo


class Ramirez2005:
    r"""
    Relation between effective temperature and color given by Ramirez and Melendez 2005.

    Ramirez and Melendez 2015, ApJ 626, 465-485 (please not that Ramirez has a non-ASCII
    accent on the i and Melendez an accent on the second e) give a relation between the
    stellar effective temperature and the color concerning various bands. This class
    allows to carry out the conversion in both directions.
    """

    def _extractTableData(self, lines, tableno):
        r"""
        Extract lines pertaining to specified table.

        Parameters
        ----------
        lines : list of strings
            Content of the data file.
        tableno : int
            Number of the table to be extracted.

        Returns
        -------
        Table data : list of strings
            Part of the file belonging to the
            specified table.
        """
        result = []
        inTable = False
        tid = "Table " + str(tableno) + ":"
        for l in lines:
            if l.find(tid) == -1:
                if not inTable:
                    continue
            else:
                inTable = True
                continue
            if inTable:
                if len(l) > 0:
                    result.append(l)
                    continue
            break
        return result

    def _convertBandName(self, bn):
        r"""
        Convert band name used in tables to internal representation.

        Parameters
        ----------
        bn : string
            Band name used in table.

        Returns
        -------
        Band ID : string
            Identifier used in the class.
        """
        name = bn.replace("{", "").replace("}", "").replace(" ", "")
        if name.startswith("("):
            name = name[1:-1]
        return name

    def _readTab23(self, lines, tableno):
        r"""
        Read tables 2 and 3.

        Parameters
        ----------
        lines : list of strings
            Content of the data file.
        tableno : int
            Number of the table to be extracted.

        Returns
        -------
        bands : list of strings
            IDs of all bands in the table.
        result : array
            Table data as array.
        """
        dat = self._extractTableData(lines, tableno)
        result = np.zeros((len(dat), 8))
        bands = [None] * len(dat)
        for i, l in enumerate(dat):
            s = l.split()
            bands[i] = self._convertBandName("".join(s[0: len(s) - 8]))
            result[i, ::] = np.array(s[-8:], dtype=float)
        return bands, result

    def _readTab45(self, lines, tableno):
        r"""
        Read tables 4 and 5.

        Parameters
        ----------
        lines : list of strings
            Content of the data file.
        tableno : int
            Number of the table to be extracted.

        Returns
        -------
        result : array
            Table data as array.
        """
        dat = self._extractTableData(lines, tableno)
        result = np.zeros((len(dat), 10))
        for i, l in enumerate(dat):
            # zero can be used instead of ldots
            l = l.replace(r"\ldots", "0.0")
            s = l.split()
            if len(s) > 10:
                # A new band identifier is given
                result[i, ::] = np.array(s[-10:], dtype=float)
            else:
                result[i, ::] = np.array(s[:], dtype=float)
        return result

    def _readData(self, fn):
        r"""
        Read the table data.

        Parameters
        ----------
        fn : string
            Filename.
        """
        # Read data and remove newlines
        lines = [s.rstrip("\n") for s in open(fn).readlines()]

        # Tables 2 and 3 have the same structure
        self._bands, self._tab2 = self._readTab23(lines, 2)
        _, self._tab3 = self._readTab23(lines, 3)
        # Tables 4 and 5 have the same structure
        self._tab4 = self._readTab45(lines, 4)
        self._tab5 = self._readTab45(lines, 5)

    def _checkBand(self, band):
        r"""
        Check whether band identifier is valid.
        """
        if not band in self._bands:
            raise (
                PE.PyAValError(
                    "No such band: " + str(band),
                    solution="Use either of: " + ", ".join(self._bands),
                )
            )

    def _checkST(self, st):
        r"""
        Check whether stellar type (main-sequence/giant) is valid.
        """
        if not st in ["ms", "g"]:
            raise (
                PE.PyAValError(
                    "No such stellar type: " + str(st),
                    solution="Either use 'ms' for main sequence or 'g' for giant.",
                )
            )

    def _resolveMetallicityIndex(self, feh):
        r"""
        Determine where to find coefficients for given metallicity in Tables 4 and 5.

        Parameters
        ----------
        feh : float
            Metallicity
        """
        fbs = [[-0.5, 0.5], [-1.5, -0.5], [-2.5, -1.5], [-4.0, -2.5]]
        for i, fb in enumerate(fbs):
            if (feh >= fb[0]) and (feh < fb[1]):
                return i
        raise (
            PE.PyAValError(
                "No data for given metallicity of " + str(feh),
                solution="See RM05, Tables 4 and 5 for ranges of validity.",
            )
        )

    def colorToTeff_nop(self, band, X, feH, stype="ms"):
        r"""
        Converts color into effective temperature according to Eq. 1.

        The conversion using to Eq. 1 neglects a polynomial correction
        for metallicity. According to RM05, this causes a systematic
        error on the order of '30 or 40 K'.

        Parameters
        ----------
        band : string
            Band identifier.
        X : float
            The color index (e.g., value of B-V).
        feH : float
            Metallicity
        stype : string, {ms, g}
            Type of star (main sequence or giant).

        Returns
        -------
        Teff : float
            The effective temperature in K.
        """
        self._checkBand(band)
        self._checkST(stype)
        if stype == "ms":
            dat = self._tab2
        else:
            dat = self._tab3

        # Row of coefficients in the table.
        j = self._bands.index(band)

        # Calculate according to Eq. 1
        result = 0.0
        for i in smo.range(3):
            result += dat[j, i] * X**i
        result += dat[j, 3] * X * feH
        for i in smo.range(4, 6):
            result += dat[j, i] * feH**i

        teff = 5040.0 / result
        return teff

    def colorToTeff(self, band, X, feH, stype="ms", ignoreRange=False):
        r"""
        Converts color into effective temperature according to Eq. 2.

        This method takes the polynomial correction into account. Note
        that no interpolation is implemented between the polynomials
        defined in Tables 4 and 5, but the appropriate polynomial
        (according to footnote (a) on under the tables) is used.

        Parameters
        ----------
        band : string
            Band identifier.
        X : float
            The color index (e.g., value of B-V).
        feH : float
            Metallicity
        stype : string, {ms, g}
            Type of star (main sequence or giant).
        ignoreRange : boolean, optional
            If True, the validity range of the relations will
            be ignored. Otherwise (default) an exception will
            be raised when a value outside the range is
            encountered.

        Returns
        -------
        Teff : float
            The effective temperature in K.
        """
        self._checkBand(band)
        self._checkST(stype)
        if stype == "ms":
            dat = self._tab2
            datp = self._tab4
        else:
            dat = self._tab3
            datp = self._tab5

        # Row of coefficients in the table.
        j = self._bands.index(band)

        # Index where to find coefficients for polynomial
        # correction in tables 4 and 5. The factor 4 occurs,
        # because there are 4 distinct metallicity bands.
        mj = j * 4 + self._resolveMetallicityIndex(feH)

        # Check range of validity
        if (X < datp[mj, 1] or X > datp[mj, 2]) and (not ignoreRange):
            raise (
                PE.PyAValError(
                    "The given color index of "
                    + str(X)
                    + " is outside the validity range of "
                    + str(datp[mj, 1])
                    + " - "
                    + str(datp[mj, 2]),
                    where="colorToTeff",
                    solution="Use index within valid range.",
                )
            )

        # Calculate according to Eq. 1
        result = 0.0
        for i in smo.range(3):
            result += dat[j, i] * X**i
        result += dat[j, 3] * X * feH
        for i in smo.range(4, 6):
            result += dat[j, i] * feH ** (i - 3)

        teff = 5040.0 / result

        # Polynomial correction
        for i in smo.range(7):
            teff += datp[mj, 3 + i] * X**i

        return teff

    def teffToColor(self, band, teff, feH, stype="ms", dteff=0.01, maxiter=100):
        r"""
        Converts effective temperature into color according to Eq. 2.

        This method inverts Eq. 2 using an iterative scheme.

        Parameters
        ----------
        band : string
            Band identifier.
        teff : float
            Effective temperature in K.
        feH : float
            Metallicity
        stype : string, {ms, g}, optional
            Type of star (main sequence or giant).
        dteff : float, optional
            Temperature difference to be reached
            by the iteration [K]. Default is 0.01.
        maxiter : int, optional
            The maximum number of iterations to be
            carried out. Default is 100.

        Returns
        -------
        X : float
            Color in the specified band.
        """
        self._checkBand(band)
        self._checkST(stype)

        # Trial step-width in x (=color) to find iterative solution
        dx = 0.1

        # Obtain guess solution neglecting polynomial
        x0 = self.teffToColor_nop(band, teff, feH, stype)
        # Iteratively improve solution
        counter = 0
        while True:
            counter += 1
            tc0 = self.colorToTeff(band, x0, feH, stype, ignoreRange=True)
            tc1 = self.colorToTeff(band, x0 + dx, feH, stype, ignoreRange=True)
            dtdx = (tc1 - tc0) / dx
            dt = tc0 - teff

            if abs(dt) < dteff:
                break
            else:
                x0 -= dt / dtdx

            if counter > maxiter:
                raise (
                    PE.PyAValError(
                        "Maxiter reached. Could not convert teff into color.",
                        where="teffToColor",
                        solutions="Try to lower precision via adjusting 'dteff'.",
                    )
                )

        return x0

    def teffToColor_nop(self, band, teff, feH, stype="ms", noRaise=False):
        r"""
        Converts effective temperature into color according to Eq. 1.

        This method inverts Eq. 1. Note that the equation is parabolic
        in the color (i.e., X). Therefore, there are two solutions of
        which the one falling within the validity ranges specified in
        Tables 4 and 5 of RM05 is selected. If none or both of the
        solutions are valid, an exception is raised.

        Parameters
        ----------
        band : string
            Band identifier.
        teff : float
            Effective temperature in K.
        feH : float
            Metallicity
        stype : string, {ms, g}
            Type of star (main sequence or giant).
        noRaise : boolean, optional
            If True, no exceptions will be raised, but warnings
            will be given Both candidate solutions will be
            returned in this case.

        Returns
        -------
        X : float
            Color in the specified band.
        """
        self._checkBand(band)
        self._checkST(stype)
        if stype == "ms":
            dat = self._tab2
            datp = self._tab4
        else:
            dat = self._tab3
            datp = self._tab5

        # Use warnings of 'noRaise' is True
        if noRaise:
            raiser = PE.warn
        else:

            def throw(x):
                raise (x)

            raiser = throw

        # Row of coefficients in the table.
        j = self._bands.index(band)

        c = dat[j, 0] + dat[j, 4] * feH + dat[j, 5] * feH**2 - 5040.0 / teff
        a13p = dat[j, 1] + dat[j, 3] * feH
        a2 = dat[j, 2]

        sq = np.sqrt(a13p**2 - 4.0 * a2 * c)
        X = [(-a13p + sq) / (2.0 * a2), (-a13p - sq) / (2.0 * a2)]

        # Get validity ranges to decide, which solution is correct
        mj = j * 4 + self._resolveMetallicityIndex(feH)

        xmin = datp[mj, 1]
        xmax = datp[mj, 2]
        xin = [False] * 2
        for i in smo.range(2):
            if X[i] >= xmin and X[i] <= xmax:
                xin[i] = True
        if sum(xin) == 2:
            raiser(
                PE.PyAAlgorithmFailure(
                    "No unique solution in inversion of teff-color relation.",
                    where="teffToColor_nop",
                    solution="Consider carefully(!) using 'noRaise'.",
                )
            )
            # Only relevant if 'noRaise' is True
            return tuple(X)
        if sum(xin) == 0:
            raiser(
                PE.PyAAlgorithmFailure(
                    "No valid solution in inversion of teff-color relation.\n"
                    + "Band: "
                    + str(band)
                    + ", Range of validity: %.4e - %.4e" % (xmin, xmax)
                    + ", candidate solutions: %.4e and %.4e" % tuple(X)
                    + ", Input Teff: %5.1f" % teff,
                    where="teffToColor_nop",
                    solution="Consider carefully(!) using 'noRaise'.",
                )
            )
            # Only relevant if 'noRaise' is True
            return tuple(X)

        if xin[0]:
            return X[0]
        else:
            return X[1]

    def availableBands(self):
        r"""
        Get a list of available band identifiers.

        Returns
        -------
        Band IDs : list of strings
            All strings used to identify bands.
        """
        return self._bands[:]

    def __init__(self):
        self._readData(os.path.join(os.path.dirname(
            __file__), "ramirez2005tables.dat"))



from __future__ import print_function, division
import os
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyaC import _ic
from .pyaConfig import *
import six.moves.urllib as urllib
from six.moves.urllib.error import URLError
import glob
from PyAstronomy.pyaC.pyaPermanent import pyaRC
 
if _ic.check["ssl"]:
    import ssl

class PyAFS:
    """
    Manage access to PyA's data directory.

    This class provides a convenient interface to create,
    read, and modify files in PyA's data directory.
    """

    def _analyzeFilename(self, fn, pedantic=True):
        """
        Split and check file or directory name

        Parameters
        ----------
        fn : string
            The file or directory name. Either absolute or relative
            to PyA's data path.
        pedantic : boolean, optional
            If True, the method will check whether the file is
            located within PyA's data path and raise an exception
            otherwise. Default is True.

        Returns
        -------
        Components : dictionary
            A dictionary containing "basename", "relPath", "absPath",
            and "fullname", where "relPath" is the path relative to
            PyA's data directory. "fullname" contains the total
            filename with absolute path. 
        """
        path, basename = os.path.split(fn)
        path = os.path.normpath(path)
        isAbsPath = os.path.isabs(path)
        if not isAbsPath:
            fullpath = os.path.join(self.dpath, path)
        else:
            fullpath = path
        fullpath = os.path.normpath(fullpath)
        if not fullpath.startswith(self.dpath) and pedantic:
            raise(PE.PyAValError("PyAFS has been asked to access the file '" + fn + "', which is not located in " +
                                 "a subdirectory of PyA's data path. Forbidden access.",
                                 solution="Use another method to access the file or check whether name is correct."))
        relativePath = os.path.relpath(fullpath, self.dpath)
        fullname = os.path.join(fullpath, basename)
        return {"basename": basename, "relPath": relativePath, "absPath": fullpath, "fullname": fullname}

    def createSubfolder(self, folder):
        """
        Create subfolders (relative to data root directory).

        Ignores already existing folders. This method can handle
        recursive directory creation.
        """
        # Assume it is a path. Append path separator
        cfold = self._analyzeFilename(folder + os.sep)
        if os.path.isdir(cfold["absPath"]):
            return
        os.makedirs(cfold["absPath"])

    def composeFilename(self, rfn):
        """
        Returns full filename.

        Parameters
        ----------
        rfn : string
            Relative file name (in PyA data path).

        Returns
        -------
        filename : string
            Absolute filename.
        """
        return os.path.join(self.dpath, rfn)

    def removeFile(self, fn):
        """
        Remove a file from PyA's data directory.

        If the file does not exist, the request is ignored.

        Parameters
        ----------
        fn : string
            The filename either absolute or relative.
        """
        ana = self._analyzeFilename(fn)
        if os.path.isfile(ana["fullname"]):
            os.remove(ana["fullname"])

    def fileExists(self, fn):
        """
        Check whether file exists

        Parameters
        ----------
        fn : string
            The filename either relative to PyA's data path or
            absolute.

        Returns
        -------
        Flag : boolean
            True if file exists and False otherwise.
        """
        ana = self._analyzeFilename(fn, False)
        return os.path.isfile(ana["fullname"])

    def _checkOnline(self, url="http://www.google.com", raiseNOC=True):
        """
        Check whether network can be reached.

        Parameters
        ----------
        url : string, optional
            The reference URL tried to be reached to check
            network availability.
        raiseNOC : boolean, optional
            If True (default), raise an exception if network
            cannot be reached.

        Returns
        -------
        Online flag : boolean
            True if network could be reached.
        """
        if not pyaRC.supposedOnline():
            PE.warn(PE.PyANetworkError("Internet connection disallowed by pyaRC.", \
                                       solution="Use 'goOnline' method in pyaRC."))
            return False
        
        online = True
        try:
            _ = urllib.request.urlopen(url)
        except Exception as e:
            # No connection seems to be available
            online = False
            if raiseNOC:
                raise(PE.PyANetworkError("You seem to be offline. Could not reach URL: '" + str(url) + "'.",
                                         solution="Get online",
                                         tbfe=e))
        return online

    def _checkContext(self, url="http://www.google.com", context=None, raiseOther=True):
        """
        Check whether `context` parameter is available in urllib.

        Parameters
        ----------
        url : string, optional
            The reference URL tried to be reached to check
            network availability.
        context : context, optional
            Context used in attempt. Default is None.
        raiseOther : boolean, optional
            If True (default), an exception is raised if an error other
            than the anticipated TypeError is raised.
        """
        contextWorks = True
        try:
            _ = urllib.request.urlopen(url, context=context)
        except TypeError as e:
            # Context appears not to work
            contextWorks = False
        except Exception as e:
            # Another error occurred
            if raiseOther:
                raise(PE.PyANetworkError("Unknown network error occurred.",
                                         tbfe=e))
        return contextWorks

    def downloadToFile(self, url, fn, clobber=False, verbose=True, openMethod=open, context=None):
        """
        Download content from URL.

        Parameters
        ----------
        url : string
            The location of the content.
        fn : string
            The relative or absolute name of the file
            to which the download shall be saved.
        clobber : boolean, optional
            If True, an existing file will be overwritten.
        verbose : boolean, optional
            If True, information on the download will be
            printed to the screen.
        openMethod : callable
            The method used to open the file to write to (default is
            open, other choices may be gzip.open or io.ipen)
        context : ssl context
            SSL context parameter handed to urlopen.
        """
        if self.fileExists(fn) and (clobber == False):
            return
        
        if not pyaRC.supposedOnline():
            PE.warn(PE.PyANetworkError("Internet connection disallowed by pyaRC.", \
                                       solution="Use 'goOnline' method in pyaRC."))
            return

        def download(url, context, nocontext=False):
            try:
                if not nocontext:
                    # Use context
                    response = urllib.request.urlopen(url, context=context)
                else:
                    # Disregard context
                    response = urllib.request.urlopen(url)
                data = response.read()     # a `bytes` object
                self.requestFile(fn, 'wb', openMethod).write(data)
            except UnicodeDecodeError as e:
                pr = urllib.parse.urlparse(url)
                if pr.scheme == "ftp":
                    # FTP fallback to circumvent default utf-8 encoding if not applicable
                    print("    (Entering fallback for FTP/latin-1 servers)")
                    from ftplib import FTP
                    ftp = FTP(pr.netloc, encoding="latin-1")
                    ftp.login()
                    with self.requestFile(fn, 'wb', openMethod) as ofi:
                        ftp.retrbinary("RETR "+pr.path, ofi.write)
                else:
                    raise e

        ana = self._analyzeFilename(fn, True)
        self.touchFile(ana["fullname"])
        try:
            if verbose:
                print("PyA download info:")
                print("  - Downloading from URL: " + str(url))
            download(url, context)
        except (KeyboardInterrupt, SystemExit):
            self.removeFile(ana["fullname"])
            raise
        except TypeError as e:
            # Possibly, context is not supported
            cs = self._checkContext()
            if not cs:
                # Network is all right, but context parameter must
                # not be specified.
                if verbose:
                    print("PyA download info:")
                    print("  - Downloading from URL: " +
                          str(url) + ", (no context)")
                download(url, context, nocontext=True)
        except URLError as e:
            # Trying to download without context (side-lining ssl verification!)
            if verbose:
                print("PyA download info:")
                print("  - Downloading from URL: " +
                      str(url) + ", (unverified context)")
            context = ssl._create_unverified_context()
            download(url, context)            
        except Exception as e:
            self.removeFile(ana["fullname"])
            sols = ["Check whether URL exists and is spelled correctly."]
            # Check whether network can be reached.
            netreach = self._checkOnline(raiseNOC=False)
            if not netreach:
                sols.append(
                    "Network could not be reached. Check your network status.")
            raise(PE.PyADownloadError("Could not download data from URL: " + str(url) + ".\n",
                                      solution=sols,
                                      tbfe=e,
                                      addInfo="Could network be reached (online)? " + {True: "yes", False: "No"}[netreach]))
        if verbose:
            print("  - Downloaded " +
                  str(os.path.getsize(ana["fullname"]) / 1000.0) + " kb")
            print("    to file: " + ana["fullname"])

    def requestFile(self, relName, mode='r', openMethod=open, *args, **kwargs):
        """
        Obtain a file object within the PyA data path.

        This method opens and creates files within PyA's data path.
        If the directory in which a file shall be created does not
        yet exist, it will be created.

        Parameters
        ----------
        relName : string
            The filename. Usually, it will be given relative to the PyA
            data path, but it can also be an absolute filename.
        mode : string, optional
            The opening mode (e.g., "r" or "w"). This flag is given to
            the `openMethod`. The default is "r" for "read".
        openMethod : method, optional
            The method used to create the file object. The default
            is Python's built-in `open` method. Another example
            could be `gzip.open`.

        Returns
        -------
        The file object.
        """
        writingMode = (('w' in mode) or ('a' in mode))
        ana = self._analyzeFilename(relName, pedantic=writingMode)
        fullName = ana["fullname"]
        if writingMode:
            self.createSubfolder(ana["absPath"])
        try:
            fileObject = openMethod(fullName, mode=mode, *args, **kwargs)
        except IOError as ioe:
            raise(PE.PyAValError("Could not access file (relative name): " + relName + ", full name: " + fullName + "\n" +
                                 "Caught exception: " + str(ioe)))
        return fileObject

    def touchFile(self, relName):
        """
        Create an empty file in PyA's data directory.

        Parameters
        ----------
        relName : string
            The filename. Usually, it will be given relative to the PyA
            data path, but it can also be an absolute filename.
        """
        fullName = self._analyzeFilename(relName)["fullname"]
        if not os.path.isfile(fullName):
            f = self.requestFile(fullName, 'w')
            f.close()
        return

    def getFullPath(self, rpath):
        """
        Return full path for relative path
        
        Parameters
        ----------
        rpath : string or list of strings
            Relative path in PyA data directory or individual
            constituents of path
        
        Returns
        -------
        Path : string
            Absolute path
        """
        return os.path.join(self.dpath, rpath)

    def globglob(self, rpath, ge):
        """
        Apply glob.glob on relative path
        
        Parameters
        ----------
        rpath : string
            Relative path
        ge : string
            Glob expression (e.g., *.*)
        
        Returns
        -------
        Fns : list of strings
            Glob results (relative paths)
        """
        fns = glob.glob(os.path.join(self.dpath, rpath, ge))
        fns = [os.path.relpath(p, start=self.dpath) for p in fns]
        return fns

    def __init__(self):
        self.conf = PyAConfig()
        # The next call will throw exception, if there is
        # no valid date directory.
        self.dpath = self.conf.getDataRoot()



from PyAstronomy.pyasl import KeplerEllipse
from PyAstronomy import funcFit as fuf
from numpy import pi
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE


class KeplerEllipseModel(fuf.OneDFit):
    """
    A model of a Keplerian orbit.

    This class uses the *KeplerEllipse* from the PyA's pyasl
    to calculate a Keplerian orbit. It may be used to fit
    complete 3d position or velocity information on the orbit;
    any individual axes may also be selected.

    The constructor allows to specify *relevant axes*, which
    are those axes considered in the calculation. The actual
    (technical) model is, however, only one-dimensional. The values
    returned by evaluate have the order
    a1, b1, c1, a2, b2, c3, ... . Where a, b, and c represent
    the first, second, and third axis and the number specifies
    the data point. Note that in this case, the resulting
    model has not the same number of points as the time
    axis.

    *Fit parameters*
      -  `a`    - The semi-major axis (same units as the data)
      - `per`   - The period (same time units as data)
      - `e`     - The eccentricity
      - `tau`   - Time of periapsis passage (same time units as data)
      - `Omega` - Longitude of the ascending node [deg]
      - `w`     - Argument of periapsis [deg]
      - `i`     - Inclination angle [deg]

    Parameters
    ----------
    relevantAxes : string
        A string containing any combination of x, y, and z.
        The string specifies the axes (and their order) to
        be considered in the calculations.
    mode : string, {"pos", "vel"}
        Determines whether the output is positions or
        velocities. In this case, the units are determined
        by the units of major axis and time (e.g., AU per day).
    """

    def __init__(self, relevantAxes="xyz", mode="pos"):
        self.ke = KeplerEllipse(1.0, 1.0)
        fuf.OneDFit.__init__(self, ["a", "per", "e", "tau", "Omega", "w", "i"])
        self["a"] = 1.0
        self["per"] = 1.0
        # Which axes to consider?
        # x=0, y=1, z=2
        self.axes = ()
        for i, axis in enumerate("xyz"):
            if axis in relevantAxes:
                self.axes += (i,)
        # Save the mode
        if not mode in ["pos", "vel"]:
            raise(PE.PyAValError("Unknown mode: " + str(mode), \
                                 where="KeplerEllipseModel", \
                                 solution="Choose either 'pos' or 'vel'."))
        self._mode = mode

    def evaluate(self, t):
        """
        Calculates and returns model according to the
        current parameter values.

        Although more than one axis may be relevant
        the output will be one dimensional. If, e.g.,
        the relevant axes are x and y, the order of
        the output will be x0, y0, x1, y1, ... .

        Parameters
        ----------
        t : array
            Times at which to evaluate the model.
        """
        self.ke.i = self["i"]
        self.ke.w = self["w"]
        self.ke.Omega = self["Omega"]
        self.ke.e = abs(self["e"])
        self.ke.a = self["a"]
        self.ke.per = self["per"]
        self.ke.tau = self["tau"]
        self.ke.n = 2.0 * pi / self["per"]
        # Get the data pertaining to the relevant axes
        if self._mode == "pos":
            result = self.ke.xyzPos(t)[::, self.axes]
        else:
            result = self.ke.xyzVel(t)[::, self.axes]
        # Reshape to 1d form
        # If the relevant axes are x and y, the order
        # of the output will be x0, y0, x1, y1, x2, y2, ...
        result = result.reshape(result.size)
        return result




class KeplerRVModel(fuf.OneDFit):
    """
    A model of a Keplerian orbit.

    This class uses the *KeplerEllipse* from the PyA's pyasl
    to calculate radial velocities for a Keplerian orbit. All calculations are
    based on the negligible companion mass hypothesis (i.e., this model has been
    implemented with planetary systems in mind).
    
    .. note:: Any planet with zero period will be ignored.

    .. note:: ? is a placeholder for an integer larger zero, indicating the number of
              the planet (the total number is controlled by the `mp` keyword).
              
    .. note:: RVs should be given in m/s and time stamps in days

    *Fit parameters*

      - `per?`   - The period in days
      - `e?`     - The eccentricity
      - `tau?`   - Time of periapsis passage 
      - `w?`     - Argument of periapsis [deg]
      - `K?`     - Semi-amplitude of radial velocity [m/s]
      - `mstar` - Stellar mass in solar masses. This parameter is usually not fitted.
                  It may be used to take into account the uncertainty on stellar mass
                  in Bayesian (MCMC) analysis.

    *Derived parameters (not to be fitted)*
      -  `a?`        - The semi-major axis in AU
      -  `MA?`       - Mean anomaly corresponding to time of first data point [deg]
      -  `msini?`    - Minimum mass (msini) in Jupiter masses 

    Parameters
    ----------
    mp : int, optional
        The number of planets considered in the model. Default is one. Note that
        signals are added, i.e., no interaction is taken into account.
    deg : int, optional
        Default is zero (i.e., a constant). The degree of a polynomial used to represent
        a systematic (non-periodic) evolution in the data.
    msun : float, optional
        Solar mass [kg]
    mJ : float, optional
        Jupiter mass [kg]
    au : float, optional
        Astronomical unit [m]
    """

    def __init__(self, mp=1, deg=0, msun=1.988547e30, mJ=1898.6e24, au=1.49597870700e11):
        self._msun = msun
        self._mJ = mJ
        self._au = au
        self._deg = deg
        self._mp = mp
        
        self.poly = fuf.PolyFit1d(deg)
        # Kepler models for individual planets
        self.kems = [KeplerEllipseModel(relevantAxes="z", mode="vel") for _ in range(1,mp+1)]
        
        pars = []
        for m in range(1, mp+1):
            # Independent parameters
            pars.extend([pn+str(m) for pn in ["K", "per", "e", "tau", "w"]])
            # Dependent parameters
            pars.extend([pn+str(m) for pn in ["msini", "a", "MA"]])
        # There is only one stellar mass
        pars.append("mstar")
        # Use parameters from polynomial
        pars.extend(self.poly.availableParameters())
        
        fuf.OneDFit.__init__(self, pars)
        self.setRootName("KeplerRVModel")
        # Dummy SMA and inclination
        for m in range(mp):
            self.kems[m]["a"] = 1.0
            self.kems[m]["i"] = 90.0
        # Default stellar mass = 0 solar mass (force thoughtful input)
        self["mstar"] = 0.0
        self.setRestriction({"mstar":[0,None]})
        
        for m in range(1, 1+mp):
            self.setRestriction({"K"+str(m):[0.,None], "per"+str(m):[0.,None], "e"+str(m):[0,1]})
    
    def _dtr(self, d):
        """ Convert degrees into rad """
        return d/180.*np.pi
    
    def evaluate(self, t):
        if self["mstar"] == 0.0:
            raise(PE.PyAValError("Stellar mass has to be specified before evaluation.", \
                                 where="KeplerRVModel"))
        
        # Take into account polynomial
        for i in range(self._deg+1):
            p = "c" + str(i)
            self.poly[p] = self[p]
        rvnew = self.poly.evaluate(t)
        
        for m in range(1, self._mp+1):
            # Name add
            nadd = str(m)
            if self["per"+nadd] == 0.0:
                continue
            for p in ["per", "e", "tau", "w"]:
                # Collect parameters from current model
                self.kems[m-1][p] = self[p+nadd]
        
                rvmodel = self.kems[m-1].evaluate(t)
            # Prevent dividing by zero
            imax = np.argmax(np.abs(rvmodel))
            cee = np.cos(self.kems[m-1].ke.trueAnomaly(t[imax]) + self._dtr(self.kems[m-1]["w"])) \
                  + self["e"+nadd]*np.cos(self._dtr(self["w"+nadd]))
            rvnew += rvmodel / rvmodel[imax] * cee * self["K"+nadd] 

            self["msini"+nadd] = self._getmsini(nadd)
            self["MA"+nadd] = self._MA(m)
            self["a"+nadd] = self._geta(nadd)
        
        return rvnew
                                                                                        
    def _getmsini(self, nadd):
        """
        Get msini
        """
        msini = self["K"+nadd] * ((self["per"+nadd]*86400.) * (self["mstar"]*self._msun)**2 / (2.*np.pi*6.67408e-11))**(1./3.) \
                * np.sqrt(1. - self["e"+nadd]**2)
        return msini/self._mJ
    
    def _MA(self, m):
        """
        Get mean anomaly corresponding to first data point [deg]
        """
        return self.kems[m-1].ke.meanAnomaly(0.0)/np.pi*180. % 360.0
    
    def _geta(self, nadd):
        """
        Get SMA in AU
        """
        return ((self["per"+nadd]*86400.)**2 * 6.67408e-11 * (self["mstar"] * self._msun) / (4.*np.pi**2) )**(1./3.) / self._au


import numpy as np

def getAngDist(ra1, dec1, ra2, dec2):
  """
    Calculate the angular distance between two coordinates.
    
    Parameters
    ----------
    ra1 : float, array
        Right ascension of the first object in degrees.
    dec1 : float, array
        Declination of the first object in degrees.
    ra2 : float, array
        Right ascension of the second object in degrees.
    dec2 : float, array
        Declination of the second object in degrees.

    Returns
    -------
    Angle : float, array
        The angular distance in DEGREES between the first
        and second coordinate in the sky.
         
  """  
  
  delt_lon = (ra1 - ra2)*np.pi/180.
  delt_lat = (dec1 - dec2)*np.pi/180.
  # Haversine formula
  dist = 2.0*np.arcsin( np.sqrt( np.sin(delt_lat/2.0)**2 + \
         np.cos(dec1*np.pi/180.)*np.cos(dec2*np.pi/180.)*np.sin(delt_lon/2.0)**2 ) )  

  return dist/np.pi*180.



from __future__ import print_function, division
import numpy as np
from .astroTimeLegacy import premat, daycnv, precess, helio_jd
from .idlMod import idlMod
from PyAstronomy.pyaC import pyaErrors as PE
import six
import six.moves as smo


def baryvel(dje, deq):
    """
    Calculate helio- and barycentric velocity.

    .. note:: The "JPL" option present in IDL is not provided here.

    Parameters
    ----------
    dje : float
        Julian ephemeris date
    deq : float
        Epoch of mean equinox of helio- and barycentric velocity output.
        If `deq` is zero, `deq` is assumed to be equal to `dje`.

    Returns
    -------
    dvelh : array
        Heliocentric velocity vector [km/s].
    dvelb : array
        Barycentric velocity vector [km/s].

    Notes
    -----

    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    pro baryvel, dje, deq, dvelh, dvelb, JPL = JPL

    NAME:
        BARYVEL  
    PURPOSE:    
        Calculates heliocentric and barycentric velocity components of Earth.

    EXPLANATION:
      BARYVEL takes into account the Earth-Moon motion, and is useful for 
      radial velocity work to an accuracy of  ~1 m/s.

    CALLING SEQUENCE:
      BARYVEL, dje, deq, dvelh, dvelb, [ JPL =  ] 

    INPUTS:
      DJE - (scalar) Julian ephemeris date.
      DEQ - (scalar) epoch of mean equinox of dvelh and dvelb. If deq=0
              then deq is assumed to be equal to dje.
    OUTPUTS: 
      DVELH: (vector(3)) heliocentric velocity component. in km/s 
      DVELB: (vector(3)) barycentric velocity component. in km/s

      The 3-vectors DVELH and DVELB are given in a right-handed coordinate 
      system with the +X axis toward the Vernal Equinox, and +Z axis 
      toward the celestial pole.      

    OPTIONAL KEYWORD SET:
      JPL - if /JPL set, then BARYVEL will call the procedure JPLEPHINTERP
            to compute the Earth velocity using the full JPL ephemeris.   
            The JPL ephemeris FITS file JPLEPH.405 must exist in either the 
            current directory, or in the directory specified by the 
            environment variable ASTRO_DATA.   Alternatively, the JPL keyword
            can be set to the full path and name of the ephemeris file.
            A copy of the JPL ephemeris FITS file is available in
                http://idlastro.gsfc.nasa.gov/ftp/data/         
    PROCEDURES CALLED:
      Function PREMAT() -- computes precession matrix
      JPLEPHREAD, JPLEPHINTERP, TDB2TDT - if /JPL keyword is set
    NOTES:
      Algorithm taken from FORTRAN program of Stumpff (1980, A&A Suppl, 41,1)
      Stumpf claimed an accuracy of 42 cm/s for the velocity.    A 
      comparison with the JPL FORTRAN planetary ephemeris program PLEPH
      found agreement to within about 65 cm/s between 1986 and 1994

      If /JPL is set (using JPLEPH.405 ephemeris file) then velocities are 
      given in the ICRS system; otherwise in the FK4 system.   
    EXAMPLE:
      Compute the radial velocity of the Earth toward Altair on 15-Feb-1994
         using both the original Stumpf algorithm and the JPL ephemeris

      IDL> jdcnv, 1994, 2, 15, 0, jd          ;==> JD = 2449398.5
      IDL> baryvel, jd, 2000, vh, vb          ;Original algorithm
              ==> vh = [-17.07243, -22.81121, -9.889315]  ;Heliocentric km/s
              ==> vb = [-17.08083, -22.80471, -9.886582]  ;Barycentric km/s
      IDL> baryvel, jd, 2000, vh, vb, /jpl   ;JPL ephemeris
              ==> vh = [-17.07236, -22.81126, -9.889419]  ;Heliocentric km/s
              ==> vb = [-17.08083, -22.80484, -9.886409]  ;Barycentric km/s

      IDL> ra = ten(19,50,46.77)*15/!RADEG    ;RA  in radians
      IDL> dec = ten(08,52,3.5)/!RADEG        ;Dec in radians
      IDL> v = vb[0]*cos(dec)*cos(ra) + $   ;Project velocity toward star
              vb[1]*cos(dec)*sin(ra) + vb[2]*sin(dec) 

    REVISION HISTORY:
      Jeff Valenti,  U.C. Berkeley    Translated BARVEL.FOR to IDL.
      W. Landsman, Cleaned up program sent by Chris McCarthy (SfSU) June 1994
      Converted to IDL V5.0   W. Landsman   September 1997
      Added /JPL keyword  W. Landsman   July 2001
      Documentation update W. Landsman Dec 2005
    """

    # Define constants
    dc2pi = 2 * np.pi
    cc2pi = 2 * np.pi
    dc1 = 1.0
    dcto = 2415020.0
    dcjul = 36525.0                     # days in Julian year
    dcbes = 0.313
    dctrop = 365.24219572               # days in tropical year (...572 insig)
    dc1900 = 1900.0
    AU = 1.4959787e8

    # Constants dcfel(i,k) of fast changing elements.
    dcfel = [1.7400353e00, 6.2833195099091e02,  5.2796e-6, 6.2565836e00, 6.2830194572674e02, -2.6180e-6, 4.7199666e00, 8.3997091449254e03, -1.9780e-5, 1.9636505e-1, 8.4334662911720e03, -5.6044e-5,
             4.1547339e00, 5.2993466764997e01,  5.8845e-6, 4.6524223e00, 2.1354275911213e01,  5.6797e-6, 4.2620486e00, 7.5025342197656e00,  5.5317e-6, 1.4740694e00, 3.8377331909193e00,  5.6093e-6]

    dcfel = np.resize(dcfel, (8, 3))

    # constants dceps and ccsel(i,k) of slowly changing elements.
    dceps = [4.093198e-1, -2.271110e-4, -2.860401e-8]
    ccsel = [1.675104e-2, -4.179579e-5, -1.260516e-7, 2.220221e-1,  2.809917e-2,  1.852532e-5, 1.589963e00,  3.418075e-2,  1.430200e-5, 2.994089e00,  2.590824e-2,  4.155840e-6, 8.155457e-1,  2.486352e-2,  6.836840e-6, 1.735614e00,  1.763719e-2,  6.370440e-6, 1.968564e00,  1.524020e-2, -2.517152e-6, 1.282417e00,  8.703393e-3,  2.289292e-5, 2.280820e00,
             1.918010e-2,  4.484520e-6, 4.833473e-2,  1.641773e-4, -4.654200e-7, 5.589232e-2, -3.455092e-4, -7.388560e-7, 4.634443e-2, -2.658234e-5,  7.757000e-8, 8.997041e-3,  6.329728e-6, -1.939256e-9, 2.284178e-2, -9.941590e-5,  6.787400e-8, 4.350267e-2, -6.839749e-5, -2.714956e-7, 1.348204e-2,  1.091504e-5,  6.903760e-7, 3.106570e-2, -1.665665e-4, -1.590188e-7]
    ccsel = np.resize(ccsel, (17, 3))

    # Constants of the arguments of the short-period perturbations.
    dcargs = [5.0974222e0, -7.8604195454652e2, 3.9584962e0, -5.7533848094674e2, 1.6338070e0, -1.1506769618935e3, 2.5487111e0, -3.9302097727326e2, 4.9255514e0, -5.8849265665348e2, 1.3363463e0, -5.5076098609303e2, 1.6072053e0, -5.2237501616674e2, 1.3629480e0, -
              1.1790629318198e3, 5.5657014e0, -1.0977134971135e3, 5.0708205e0, -1.5774000881978e2, 3.9318944e0,  5.2963464780000e1, 4.8989497e0,  3.9809289073258e1, 1.3097446e0,  7.7540959633708e1, 3.5147141e0,  7.9618578146517e1, 3.5413158e0, -5.4868336758022e2]
    dcargs = np.resize(dcargs, (15, 2))

    # Amplitudes ccamps(n,k) of the short-period perturbations.
    ccamps = \
        [-2.279594e-5,  1.407414e-5,  8.273188e-6,  1.340565e-5, -2.490817e-7, -3.494537e-5,  2.860401e-7,  1.289448e-7,  1.627237e-5, -1.823138e-7, 6.593466e-7,  1.322572e-5,  9.258695e-6, -4.674248e-7, -3.646275e-7, 1.140767e-5, -2.049792e-5, -4.747930e-6, -2.638763e-6, -1.245408e-7, 9.516893e-6, -2.748894e-6, -1.319381e-6, -4.549908e-6, -1.864821e-7, 7.310990e-6, -1.924710e-6, -8.772849e-7, -3.334143e-6, -1.745256e-7, -2.603449e-6,  7.359472e-6,  3.168357e-6,  1.119056e-6, -1.655307e-7, -3.228859e-6,
            1.308997e-7,  1.013137e-7,  2.403899e-6, -3.736225e-7, 3.442177e-7,  2.671323e-6,  1.832858e-6, -2.394688e-7, -3.478444e-7, 8.702406e-6, -8.421214e-6, -1.372341e-6, -1.455234e-6, -4.998479e-8, -1.488378e-6, -1.251789e-5,  5.226868e-7, -2.049301e-7,  0.e0, -8.043059e-6, -2.991300e-6,  1.473654e-7, -3.154542e-7,  0.e0, 3.699128e-6, -3.316126e-6,  2.901257e-7,  3.407826e-7,  0.e0, 2.550120e-6, -1.241123e-6,  9.901116e-8,  2.210482e-7,  0.e0, -6.351059e-7,  2.341650e-6,  1.061492e-6,  2.878231e-7,  0.e0]
    ccamps = np.resize(ccamps, (15, 5))

    # Constants csec3 and ccsec(n,k) of the secular perturbations in longitude.
    ccsec3 = -7.757020e-8
    ccsec = [1.289600e-6, 5.550147e-1, 2.076942e00, 3.102810e-5, 4.035027e00, 3.525565e-1,
             9.124190e-6, 9.990265e-1, 2.622706e00, 9.793240e-7, 5.508259e00, 1.559103e01]
    ccsec = np.resize(ccsec, (4, 3))

    # Sidereal rates.
    dcsld = 1.990987e-7                   # sidereal rate in longitude
    ccsgd = 1.990969e-7                   # sidereal rate in mean anomaly

    # Constants used in the calculation of the lunar contribution.
    cckm = 3.122140e-5
    ccmld = 2.661699e-6
    ccfdi = 2.399485e-7

    # Constants dcargm(i,k) of the arguments of the perturbations of the motion
    # of the moon.
    dcargm = [5.1679830e0,  8.3286911095275e3, 5.4913150e0, -
              7.2140632838100e3, 5.9598530e0,  1.5542754389685e4]
    dcargm = np.resize(dcargm, (3, 2))

    # Amplitudes ccampm(n,k) of the perturbations of the moon.
    ccampm = [1.097594e-1, 2.896773e-7, 5.450474e-2,  1.438491e-7, -2.223581e-2, 5.083103e-8,
              1.002548e-2, -2.291823e-8, 1.148966e-2, 5.658888e-8, 8.249439e-3,  4.063015e-8]
    ccampm = np.resize(ccampm, (3, 4))

    # ccpamv(k)=a*m*dl,dt (planets), dc1mme=1-mass(earth+moon)
    ccpamv = [8.326827e-11, 1.843484e-11, 1.988712e-12, 1.881276e-12]
    dc1mme = 0.99999696e0

    # Time arguments.
    dt = (dje - dcto) / dcjul
    tvec = np.array([1e0, dt, dt * dt])

    # Values of all elements for the instant(aneous?) dje.
    temp = idlMod(np.dot(dcfel, tvec), dc2pi)

    dml = temp[0]
    forbel = temp[1:8]
    g = forbel[0]                         # old fortran equivalence

    deps = idlMod(np.sum(tvec * dceps), dc2pi)
    sorbel = idlMod(np.dot(ccsel, tvec), dc2pi)
    e = sorbel[0]                         # old fortran equivalence

    # Secular perturbations in longitude.
    dummy = np.cos(2.0)
    sn = np.sin(idlMod(np.dot(ccsec[::, 1:3], tvec[0:2]), cc2pi))

    # Periodic perturbations of the emb (earth-moon barycenter).
    pertl = np.sum(ccsec[::, 0] * sn) + (dt * ccsec3 * sn[2])
    pertld = 0.0
    pertr = 0.0
    pertrd = 0.0
    for k in smo.range(15):
        a = idlMod((dcargs[k, 0] + dt * dcargs[k, 1]), dc2pi)
        cosa = np.cos(a)
        sina = np.sin(a)
        pertl = pertl + ccamps[k, 0] * cosa + ccamps[k, 1] * sina
        pertr = pertr + ccamps[k, 2] * cosa + ccamps[k, 3] * sina
        if k < 11:
            pertld = pertld + (ccamps[k, 1] * cosa -
                               ccamps[k, 0] * sina) * ccamps[k, 4]
            pertrd = pertrd + (ccamps[k, 3] * cosa -
                               ccamps[k, 2] * sina) * ccamps[k, 4]

    # Elliptic part of the motion of the emb.
    phi = (e * e / 4e0) * (((8e0 / e) - e) * np.sin(g) + 5 *
                           np.sin(2 * g) + (13 / 3e0) * e * np.sin(3 * g))
    f = g + phi
    sinf = np.sin(f)
    cosf = np.cos(f)
    dpsi = (dc1 - e * e) / (dc1 + e * cosf)
    phid = 2 * e * ccsgd * ((1 + 1.5 * e * e) * cosf +
                            e * (1.25 - 0.5 * sinf * sinf))
    psid = ccsgd * e * sinf / np.sqrt(dc1 - e * e)

    # Perturbed heliocentric motion of the emb.
    d1pdro = dc1 + pertr
    drd = d1pdro * (psid + dpsi * pertrd)
    drld = d1pdro * dpsi * (dcsld + phid + pertld)
    dtl = idlMod((dml + phi + pertl), dc2pi)
    dsinls = np.sin(dtl)
    dcosls = np.cos(dtl)
    dxhd = drd * dcosls - drld * dsinls
    dyhd = drd * dsinls + drld * dcosls

    # Influence of eccentricity, evection and variation on the geocentric
    # motion of the moon.
    pertl = 0.0
    pertld = 0.0
    pertp = 0.0
    pertpd = 0.0
    for k in smo.range(3):
        a = idlMod((dcargm[k, 0] + dt * dcargm[k, 1]), dc2pi)
        sina = np.sin(a)
        cosa = np.cos(a)
        pertl = pertl + ccampm[k, 0] * sina
        pertld = pertld + ccampm[k, 1] * cosa
        pertp = pertp + ccampm[k, 2] * cosa
        pertpd = pertpd - ccampm[k, 3] * sina

    # Heliocentric motion of the earth.
    tl = forbel[1] + pertl
    sinlm = np.sin(tl)
    coslm = np.cos(tl)
    sigma = cckm / (1.0 + pertp)
    a = sigma * (ccmld + pertld)
    b = sigma * pertpd
    dxhd = dxhd + a * sinlm + b * coslm
    dyhd = dyhd - a * coslm + b * sinlm
    dzhd = -sigma * ccfdi * np.cos(forbel[2])

    # Barycentric motion of the earth.
    dxbd = dxhd * dc1mme
    dybd = dyhd * dc1mme
    dzbd = dzhd * dc1mme
    for k in smo.range(4):
        plon = forbel[k + 3]
        pomg = sorbel[k + 1]
        pecc = sorbel[k + 9]
        tl = idlMod((plon + 2.0 * pecc * np.sin(plon - pomg)), cc2pi)
        dxbd = dxbd + ccpamv[k] * (np.sin(tl) + pecc * np.sin(pomg))
        dybd = dybd - ccpamv[k] * (np.cos(tl) + pecc * np.cos(pomg))
        dzbd = dzbd - ccpamv[k] * sorbel[k + 13] * np.cos(plon - sorbel[k + 5])

    # Transition to mean equator of date.
    dcosep = np.cos(deps)
    dsinep = np.sin(deps)
    dyahd = dcosep * dyhd - dsinep * dzhd
    dzahd = dsinep * dyhd + dcosep * dzhd
    dyabd = dcosep * dybd - dsinep * dzbd
    dzabd = dsinep * dybd + dcosep * dzbd

    # Epoch of mean equinox (deq) of zero implies that we should use
    # Julian ephemeris date (dje) as epoch of mean equinox.
    if deq == 0:
        dvelh = AU * np.array([dxhd, dyahd, dzahd])
        dvelb = AU * np.array([dxbd, dyabd, dzabd])
        return dvelh, dvelb

    # General precession from epoch dje to deq.
    deqdat = (dje - dcto - dcbes) / dctrop + dc1900
    prema = np.transpose(premat(deqdat, deq, FK4=True))

    dvelh = AU * np.dot([dxhd, dyahd, dzahd], prema)
    dvelb = AU * np.dot([dxbd, dyabd, dzabd], prema)
    return dvelh, dvelb


def baryCorr(jd, ra, dec, deq=0.0):
    """
    Calculate barycentric correction.

    This function uses the :py:func:`baryvel` function
    to calculate the helio- and barycentric motion of
    the Earth and projects it onto the direction to
    the star.

    .. note:: Positive return values indicate that the Earth
              moves toward the star.

    Parameters
    ----------
    jd : float
        The time at which to calculate the correction.
    ra : float
        Right ascension in degrees.
    dec : float
        Declination in degrees.
    deq : float, optional
        The mean equinox of barycentric velocity
        calculation (see :py:func:`bryvel`). If zero,
        it is assumed to be the same as `jd`.

    Returns
    -------
    Projected heliocentric velocity : float
        Heliocentric velocity toward star [km/s]
    Projected barycentric velocity : float
        Barycentric velocity toward star [km/s]
    """
    vh, vb = baryvel(jd, deq)
    rarad = ra / 180. * np.pi
    derad = dec / 180. * np.pi

    # Projected velocity toward star
    vbp = vb[0] * np.cos(derad) * np.cos(rarad) + \
        vb[1] * np.cos(derad) * np.sin(rarad) + vb[2] * np.sin(derad)
    vhp = vh[0] * np.cos(derad) * np.cos(rarad) + \
        vh[1] * np.cos(derad) * np.sin(rarad) + vh[2] * np.sin(derad)

    return vhp, vbp


def helcorr(obs_long, obs_lat, obs_alt, ra2000, dec2000, jd, debug=False):
    """
    Calculate barycentric velocity correction.

    This function calculates the motion of an observer in
    the direction of a star. In contrast to :py:func:`baryvel`
    and :py:func:`baryCorr`, the rotation of the Earth is
    taken into account.
    
    The coordinates (ra2000, dec2000) are precessed to the epoch
    defined by jd. These coordinates are used in the calculation.

    .. note:: This function was ported from the REDUCE IDL package.
              See Piskunov & Valenti 2002, A&A 385, 1095 for a detailed
              description of the package and/or visit
              http://www.astro.uu.se/~piskunov/RESEARCH/REDUCE/

    .. warning:: Contrary to the original implementation the longitude
                 increases toward the East and the right ascension is
                 given in degrees instead of hours. The JD is given as is,
                 in particular, nothing needs to be subtracted.

    Parameters
    ----------
    obs_long : float
        Longitude of observatory (degrees, **eastern** direction is positive)
    obs_lat : float
        Latitude of observatory [deg]
    obs_alt : float
        Altitude of observatory [m]
    ra2000 : float
        Right ascension of object for epoch 2000.0 [deg]
    dec2000 : float
        Declination of object for epoch 2000.0 [deg]
    jd : float
        Julian date for the middle of exposure.

    Returns
    -------
    Barycentric correction : float
        The barycentric correction accounting for the rotation
        of the Earth, the rotation of the Earth's center around
        the Earth-Moon barycenter, and the motion of the Earth-Moon 
        barycenter around the center of the Sun [km/s].
    HJD : float
        Heliocentric Julian date for middle of exposure.

    Notes
    -----

    :IDL REDUCE - Documentation:


    Calculates heliocentric Julian date, barycentric and heliocentric radial
    velocity corrections from:

    INPUT:
    <OBSLON> Longitude of observatory (degrees, western direction is positive)
    <OBSLAT> Latitude of observatory (degrees)
    <OBSALT> Altitude of observatory (meters)
    <RA2000> Right ascension of object for epoch 2000.0 (hours)
    <DE2000> Declination of object for epoch 2000.0 (degrees)
    <JD> Julian date for the middle of exposure
    [DEBUG=] set keyword to get additional results for debugging

    OUTPUT:
    <CORRECTION> barycentric correction - correction for rotation of earth,
       rotation of earth center about the earth-moon barycenter, earth-moon 
       barycenter about the center of the Sun.
    <HJD> Heliocentric Julian date for middle of exposure

    Algorithms used are taken from the IRAF task noao.astutils.rvcorrect
    and some procedures of the IDL Astrolib are used as well.
    Accuracy is about 0.5 seconds in time and about 1 m/s in velocity.

    History:
    written by Peter Mittermayer, Nov 8,2003
    2005-January-13   Kudryavtsev   Made more accurate calculation of the sidereal time.
                                    Conformity with MIDAS compute/barycorr is checked.
    2005-June-20      Kochukhov Included precession of RA2000 and DEC2000 to current epoch

    """
    from PyAstronomy.pyaC import degtorad

    # This reverts the original longitude convention. After this,
    # East longitudes are positive
    obs_long = -obs_long

    if (ra2000 < 0.0) or (ra2000 > 360.):
        raise(PE.PyAValError("Invalid value for ra2000 (value given is " + str(ra2000) + ").",
                             where="helcorr",
                             solution="Adjust right ascension (0 - 360)."))
    if (dec2000 < -90.0) or (dec2000 > 90.0):
        raise(PE.PyAValError("Invalid value for dec2000 (value given is " + str(dec2000) + ").",
                             where="helcorr",
                             solution="Adjust declination (-90 -- +90)."))

    if jd < 2.4e6:
        PE.warn(PE.PyAValError("The given Julian Date (" + str(jd) +
                               ") is exceedingly small. Did you subtract 2.4e6?"))

    # Covert JD to Gregorian calendar date
    xjd = jd

    year, month, day, ut = tuple(daycnv(xjd))

    # Current epoch
    epoch = year + month / 12. + day / 365.

    # Precess ra2000 and dec2000 to current epoch, resulting ra is in degrees
    ra = ra2000
    dec = dec2000
    ra, dec = precess(ra, dec, 2000.0, epoch)

    # Calculate heliocentric julian date
    rjd = jd - 2.4e6
    hjd = helio_jd(rjd, ra, dec) + 2.4e6

    # DIURNAL VELOCITY (see IRAF task noao.astutil.rvcorrect)
    # convert geodetic latitude into geocentric latitude to correct
    # for rotation of earth
    dlat = -(11. * 60. + 32.743) * np.sin(2.0 * degtorad(obs_lat)) \
        + 1.1633 * np.sin(4.0 * degtorad(obs_lat)) - \
        0.0026 * np.sin(6.0 * degtorad(obs_lat))
    lat = obs_lat + dlat / 3600.0

    # Calculate distance of observer from earth center
    r = 6378160.0 * (0.998327073 + 0.001676438 * np.cos(2.0 * degtorad(lat))
                     - 0.00000351 * np.cos(4.0 * degtorad(lat)) + 0.000000008 * np.cos(6.0 * degtorad(lat))) \
        + obs_alt

    # Calculate rotational velocity (perpendicular to the radius vector) in km/s
    # 23.934469591229 is the sidereal day in hours for 1986
    v = 2. * np.pi * (r / 1000.) / (23.934469591229 * 3600.)

    # Calculating local mean sidereal time (see astronomical almanach)
    tu = (rjd - 51545.0) / 36525.0
    gmst = 6.697374558 + ut + \
        (236.555367908 * (rjd - 51545.0) +
         0.093104 * tu**2 - 6.2e-6 * tu**3) / 3600.0
    lmst = idlMod(gmst - obs_long / 15., 24)

    # Projection of rotational velocity along the line of sight
    vdiurnal = v * np.cos(degtorad(lat)) * np.cos(degtorad(dec)
                                                  ) * np.sin(degtorad(ra - lmst * 15))

    # BARICENTRIC and HELIOCENTRIC VELOCITIES
    vh, vb = baryvel(xjd, 0)

    # Project to line of sight
    vbar = vb[0] * np.cos(degtorad(dec)) * np.cos(degtorad(ra)) + vb[1] * np.cos(degtorad(dec)) * np.sin(degtorad(ra)) + \
        vb[2] * np.sin(degtorad(dec))
    vhel = vh[0] * np.cos(degtorad(dec)) * np.cos(degtorad(ra)) + vh[1] * np.cos(degtorad(dec)) * np.sin(degtorad(ra)) + \
        vh[2] * np.sin(degtorad(dec))

    # Use barycentric velocity for correction
    corr = (vdiurnal + vbar)

    if debug:
        print('')
        print('----- HELCORR.PRO - DEBUG INFO - START ----')
        print(
            '(obs_long (East positive),obs_lat,obs_alt) Observatory coordinates [deg,m]: ', -obs_long, obs_lat, obs_alt)
        print(
            '(ra,dec) Object coordinates (for epoch 2000.0) [deg]: ', ra2000, dec2000)
        print(
            '(ra,dec) Object coordinates (precessed to epoch %f) [deg]: ' % (epoch), ra, dec)
        print('(ut) Universal time (middle of exposure) [hrs]: ', ut)
        print('(jd) Julian date (middle of exposure) (JD): ', jd)
        print('(hjd) Heliocentric Julian date (middle of exposure) (HJD): ', hjd)
        print('(gmst) Greenwich mean sidereal time [hrs]: ', idlMod(gmst, 24))
        print('(lmst) Local mean sidereal time [hrs]: ', lmst)
        print('(dlat) Latitude correction [deg]: ', dlat)
        print('(lat) Geocentric latitude of observer [deg]: ', lat)
        print('(r) Distance of observer from center of earth [m]: ', r)
        print(
            '(v) Rotational velocity of earth at the position of the observer [km/s]: ', v)
        print(
            '(vdiurnal) Projected earth rotation and earth-moon revolution [km/s]: ', vdiurnal)
        print('(vbar) Barycentric velocity [km/s]: ', vbar)
        print('(vhel) Heliocentric velocity [km/s]: ', vhel)
        print('(corr) Vdiurnal+vbar [km/s]: ', corr)
        print('----- HELCORR.PRO - DEBUG INFO - END -----')
        print('')

    return corr, hjd



from __future__ import print_function, division
import numpy as np
import inspect
from PyAstronomy.pyaC import pyaErrors as PE
from scipy import optimize as sco
from PyAstronomy import constants as PC


def _bisect_root_find(f, ab, nroot, eps, roots=None, bqargs={}):
    r"""
    Find roots by bisection method and scipy's brentq
    
    Parameters are the same is those of bisect_root_find.
    
    Parameters
    ----------
    ab : list of intervals
        An interval is a two-list or tuple giving the lower and
        upper limit of the interval
    """
    
    if roots is None:
        roots = []
    
    if len(roots) >= nroot:
        return roots
    
    def addroot(root, roots):
        # Add root unless too close
        if sum([abs(root-r)<eps for r in roots]) == 0:
            roots.append(root)
    
    if len(ab) == 0:
        # No new intervals (possibly all <eps)
        return roots
    
    # Prepare list of new intervals
    nab = []
    for i in ab:
        try:
            root = sco.brentq(f, i[0], i[1], **bqargs)
        except ValueError:
            # No root
            w = i[1]-i[0]
            if w > eps:
                nab.extend([[i[0], i[0]+w/2], [i[0]+w/2,i[1]]])
        else:
            # root
            addroot(root, roots)
            if root-i[0] > eps:
                nab.append([i[0], root-eps/2])
            if i[1]-root > eps:
                nab.append([root+eps/2, i[1]])
    return _bisect_root_find(f, nab, nroot, eps, roots=roots)
    
def bisect_root_find(f, a, b, nroot, eps, bqargs={}):
    r"""
    Find multiple roots using bisection method and scipy's brentq algorithm
    
    Parameters
    ----------
    f : callable
        Callable taking a single value as argument and
        returning a scalar
    a, b : float
        Lower and upper limits of the interval
    nroot: int
        Number of roots to be found. The function will not
        stop iterating before that number is found. It may
        return more than that in advantageous settings or
        less than that if the minimum interval length (eps)
        is reached before nroot roots are found.
    eps : float
        Minimum distance between roots.
    bqargs : dictionary, optional
        Additional keywords handed to scipy's brentq.
    """
    w = b-a
    nab = [[a, a+w/2], [a+w/2, b]]
    return _bisect_root_find(f, nab, nroot, eps, bqargs=bqargs)



def _r1r2_dl(x, y, z):
    r"""
    Calculate r1 and r2 for dimensionless Roche potential
    """
    r1 = np.sqrt(x**2 + y**2 + z**2)
    r2 = np.sqrt((x-1)**2 + y**2 + z**2)
    return r1, r2

def _checkq(q):
    """ Check validity of q """
    if (q < 0) or (q > 1):
        raise(PE.PyAValError("Invalid mass ratio q=m2/m1 of '" + str(q) + "'.", \
                             where=inspect.stack()[1][3], \
                             solution="Check masses. Perhaps exchange m1 and m2 so that m1 >= m2."))

def _checkm(m):
    if not m in [1,2]:
        raise(PE.PyAValError("Invalid m(ass). Current value is '" + str(m) + "'", \
                             where=inspect.stack()[1][3], \
                             solution="Use 1 or 2."))

def rochepot_dl(x, y, z, q):
    r"""
    Dimensionless Roche potential (:math:`\Phi_n`, synchronous rotation)
    
    More massive component (:math:`m_1`) is centered at (x,y,z) = (0,0,0). Less massive
    component (:math:`m_2`) is at (1,0,0). The unit of length is the distance between
    the objects. Both objects are in the x,y plane (x-axis along the connecting line and
    z perpendicular to the orbital plane). 
    
    Parameters
    ----------
    x, y, z : float or array
        Location(s) at which to calculate the potential. Unit
        of length is the distance between the masses m1 and m2.
    q : float
        Mass ratio (0 <= m2/m1 <= 1)
    
    Returns
    -------
    Potential : float or array
        The potential at the specified location(s)
    """
    _checkq(q)
    r1, r2 = _r1r2_dl(x, y, z)
    p = 2/((1+q)*r1) + 2*q/((1+q)*r2) + (x - q/(1+q))**2 + y**2
    return p

def ddx_rochepot_dl(x, q, y=0, z=0):
    r"""
    Derivative of dimensionless Roche potential along x-axis
    
    Parameters
    ----------
    x, y, z : float or array
        Position (default for y and z is zero)
    q : float
        Mass ratio m2/m1
    
    Returns
    -------
    Derivative: float or array
        d/dx of dimensionless Roche potential
    """
    _checkq(q)
    r1, r2 = _r1r2_dl(x, y, z)
    ddx = 2/(1+q) * (-x/r1**3) + 2*q/(1+q) * (-(x-1)/r2**3) + 2*(x - q/(1+q))
    return ddx

def ddz_rochepot_dl(z, q, x=1, y=0):
    r"""
    Derivative of dimensionless Roche potential along z-axis
    
    Parameters
    ----------
    x, y, z : float or array
        Position (default for x is one and zero for z)
    q : float
        Mass ratio m2/m1
    
    Returns
    -------
    Derivative: float or array
        d/dz of dimensionless Roche potential
    """
    _checkq(q)
    r1, r2 = _r1r2_dl(x, y, z)
    ddz = 2/(1+q) * (-z/r1**3) + 2*q/(1+q) * (-z/r2**3) 
    return ddz

def ddy_rochepot_dl(y, q, x=1, z=0):
    r"""
    Derivative of dimensionless Roche potential along y-axis
    
    Parameters
    ----------
    x, y, z : float or array
        Position (default for x is one and zero for z)
    q : float
        Mass ratio m2/m1
    
    Returns
    -------
    Derivative: float or array
        d/dy of dimensionless Roche potential
    """
    _checkq(q)
    r1, r2 = _r1r2_dl(x, y, z)
    ddy = 2/(1+q) * (-y/r1**3) + 2*q/(1+q) * (-y/r2**3) + 2*y 
    return ddy

def roche_lobe_radius_eggleton(q, m):
    r"""
    Approximate effective dimensionless Roche lobe radius according to Eggelton 1983 (ApJ 268, 368).
    
    The effective Roche lobe radius is the radius of a sphere with the same
    volume as the actual equipotential surface defining the Roche lobe.
    
    .. math::
    
        r_L \approx \frac{0.49\,q^{2/3}}{0.6\,q^{2/3} + \ln(1+q^{1/3})}
    
    Parameters
    ----------
    q : float or array
        Mass ratio (m2/m1)
    m : int, {1,2}
        Calculate radius for mass 1 or 2
    """
    _checkq(q)
    _checkm(m)
    if m == 1:
        # Hilditch p161
        q = 1/q
    q23 = q**(2./3)
    return 0.49*q23 / (0.6*q23 + np.log(1+q**(1./3)))

def _get_lagrange_123(q, a, b, getdlrp=True):
    r"""
    Find root of derivative along x-axis
    
    Parameters
    ----------
    q : float
        Mass ratio (m2/m1)
    a, b : float
        Limits for root search
    getdlrp : boolean, optional
        If True (default), also return potential at that point
    """
    r = sco.brentq(ddx_rochepot_dl, a, b, args=(q))
    if getdlrp:
        dlrp = rochepot_dl(r, 0, 0, q)
        return r, dlrp
    return r

def get_lagrange_1(q, getdlrp=True, eps=1e-4):
    r"""
    Get location of first Lagrange point
    
    Parameters
    ----------
    q : float
        Mass ratio
    getdlrp : boolean, optional
        If True (default), also the dimensionless Roche potential
        at the first Lagrange point is returned.
    eps : float, optional
        The potential diverges at x=0 and x=1. The search for a root of
        the derivative is performed in the interval [eps, 1-eps].
    
    Returns
    -------
    xL1 : float
        The location of the first Lagrange point (xL1, 0, 0).
    Potential : float, optional
        The dimensionless potential at the first Lagrange point (default is True).
    """
    _checkq(q)
    return _get_lagrange_123(q, eps, 1-eps, getdlrp)

def get_epradius_ss_polar_side(q, pot=None, eps=1e-6):
    r"""
    Get dimensionless radii of equipotential surface for secondary mass in three directions
    
    Assumes that the equipotential surface is enclosed within the Roche lobe
    
    Parameters
    ----------
    q : float
        Mass ratio m2/m1
    pot : float, optional
        Dimensionless potential for equipotential surface. If None (default),
        the L1 potential will be adopted. 
    eps : float, optional
        Lowest dimensionless radius considered in the
        search.
    brl : boolean, optional
        Search below Roche lobe (default is True).
    
    Returns
    -------
    rx : float
        Substellar radius, height along connecting line (dimensionless)
    rz : float
        Polar radius (along z-axis)
    ry : float
        Side radius (along y-axis)
    """
    _checkq(q)
    # x-axis (substellar point)
    rlr, l1pot = get_lagrange_1(q, getdlrp=True, eps=eps)
    if (pot is not None) and (pot < l1pot):
        raise(PE.PyAValError(f"Given potential {pot} is below Roche lobe potential {l1pot} (minimum)", \
                             where="get_epradius_ss_polar_side"))
    
    if (pot is None) or (abs(pot-l1pot)<1e-6):
        pot = l1pot
        xros = np.array([rlr])
    else:
        f = lambda x: rochepot_dl(x,0,0,q) - pot
        xros = np.array(sorted(bisect_root_find(f, eps, 1-eps, 2, 1e-2), reverse=True))
    # Positive height 
    rx = 1 - xros
    # y-axis (side radius)
    f = lambda y: rochepot_dl(1,y,0,q) - pot
    ry = sorted(np.array(sorted(bisect_root_find(f, eps, rlr, 1, 1e-4))))
    # z-axis (polar radius)
    f = lambda z: rochepot_dl(1,0,z,q) - pot
    rz = sorted(np.array(sorted(bisect_root_find(f, eps, rlr, 1, 1e-4))))
    return rx[0], rz[0], ry[0]

def get_radius_ss_polar_side(q, sma, reff, eps=1e-6):
    r"""
    Calculate substellar, polar, and side radii of secondary body
    
    The function first finds the dimensionless Roche potential corresponding
    to the equipotential surface corresponding to the effective radius and
    then determines the radii of that surface along the x, y, and z directions. 
    
    The effective radius is the radius of a sphere with the same area as
    the ellipse spanned by the polar and side radii of the equipotential
    area.
    
    Parameters
    ----------
    q : float
        Mass ratio m2/m1
    sma : float
        Semi-major axis [AU]
    reff : float
        Effective radius of the body [RJ]
    
    Returns
    -------
    rss, rpol, rside : float
        Radius of equipotential surface at substellar point (along the connecting x-axis),
        polar radius (along the z-axis), and radius toward the side (y-axis).
    pequi : float
        Dimensionless potential of equipotential surface corresponding to effective
        radius.
    """
    _checkq(q)
    pc = PC.PyAConstants()
    # Dimensionless effective radius
    reff_dl = reff*pc.RJ / (sma*pc.AU)
    # Potential at effective radius
    peff = rochepot_dl(1-reff_dl, 0, 0, q)
    
    def darea(pot):
        """ Difference between effective area and area of ellipse (polar x side) """
        _, rz, ry = get_epradius_ss_polar_side(q, pot, eps=eps)
        return rz*ry - reff_dl**2
    
    _, l1pot = get_lagrange_1(q, getdlrp=True)
    rpot = sco.brentq(darea, l1pot+1e-10, peff*10)
    
    rr = get_epradius_ss_polar_side(q, rpot, eps=eps)
    rjs = [r*sma*PC.AU/pc.RJ for r in rr]
    return *rjs, rpot
    
def get_lagrange_2(q, getdlrp=True, eps=1e-4):
    r"""
    Get location of second Lagrange point
    
    Parameters
    ----------
    q : float
        Mass ratio
    getdlrp : boolean, optional
        If True (default), also the dimensionless Roche potential
        at the first Lagrange point is returned.
    eps : float, optional
        The potential diverges at x=0 and x=1. The search for a root of
        the derivative is performed in the interval [1+eps, 1+10*rL], where rL
        is the estimated (effective) Roche lobe radius according to Eggleton 1983.
    
    Returns
    -------
    xL2 : float
        The location of the second Lagrange point (xL2, 0, 0).
    Potential : float, optional
        The dimensionless potential at the second Lagrange point (default is True).
    """
    _checkq(q)
    rl = roche_lobe_radius_eggleton(q,2)
    return _get_lagrange_123(q, 1+eps, 1+10*rl, getdlrp)

def get_lagrange_3(q, getdlrp=True, eps=1e-4):
    r"""
    Get location of third Lagrange point
    
    Parameters
    ----------
    q : float
        Mass ratio
    getdlrp : boolean, optional
        If True (default), also the dimensionless Roche potential
        at the third Lagrange point is returned.
    eps : float, optional
        The potential diverges at x=0. The search for a root of
        the derivative is performed in the interval [-10, -eps].
    
    Returns
    -------
    xL3 : float
        The location of the first Lagrange point (xL3, 0, 0).
    Potential : float, optional
        The dimensionless potential at the third Lagrange point (default is True).
    """
    _checkq(q)
    return _get_lagrange_123(q, -10, -eps, getdlrp)

def get_lagrange_4():
    r"""
    Get location of forth Lagrange point
    
    Orbital angular momentum is supposed to point into +z direction.
    L4 is in direction of motion of m2.
    
    Returns
    -------
    x,y,z : float
        The location of the fourth Lagrange point (xL3, 0, 0).
    """
    return (0.5, np.sin(np.radians(60)), 0)

def get_lagrange_5(getdlrp=True):
    r"""
    Get location of fifth Lagrange point
    
    Orbital angular momentum is supposed to point into +z direction.
    L5 is behind motion of m2.
    
    Returns
    -------
    x,y,z : float
        The location of the fifth Lagrange point (xL3, 0, 0).
    """
    return (0.5, -np.sin(np.radians(60)), 0)

def roche_yz_extent(q, m=2, pl=None, eps=1e-4):
    r"""
    Extent of equipotential surface in y and z direction
    
    Parameter
    ---------
    q : float
        Mass ratio m2/m1
    m : int, {1,2}, optional
        Whether to center on m1 or m2
    pl : float, optional
        The (dimensional) potential level to be considered. If None (default),
        the Roche lobe (L1) potential level is used.
    eps : float, optional
        Margin used to avoid singularities of Roche potential. Default is 1e-4.
    
    Returns
    -------
    dy, dz : float
        Distance to equipotential level along y and z direction.
    """
    _checkm(m)
    _checkq(q)
    x = m - 1
    l1, l1pot = get_lagrange_1(q, getdlrp=True, eps=eps)
    if m == 2:
        l1 = 1 - l1
    if pl is None:
        pl = l1pot
    f = lambda y: rochepot_dl(x,y,0,q) - pl
    p1y = sco.brentq(f, eps, l1)
    f = lambda z: rochepot_dl(x,0,z,q) - pl
    p1z = sco.brentq(f, eps, l1)
    return p1y, p1z

def roche_vol_MC(q, m=2, n=100000, pl=None, eps=1e-4, fullout=True):
    r"""
    Calculate (dimensionless) volume of equipotential surface such as the Roche lobe
    
    Uses Monte Carlo (MC) integration
    
    Parameters
    ----------
    q : float
        Mass ratio (m2/m1)
    m : int, {1,2}
        Whether to calculate volume for m1 or m2
    n : int, optional
        Number of samples to be used in the Monte Carlo integration. Default is 100000.
        Increase to improve accuracy.
    pl : float, optional
        The (dimensionless) potential level bounding the volume. If None (default), the Roche
        lobe potential (L1) is used.
    eps : float, optional
        Margin used to avoid singularities of Roche potential. Default is 1e-4.
    fullout : boolean, optional
        If True (default), provides more output than volume.
    
    Returns
    -------
    V, Verr : floats
        Dimensionless volume of Roche lobe and estimate of uncertainty
    Reff, Refferr: floats, optional
        Radius of a sphere with the same radius (dimensionless) and estimate
        of uncertainty. Provided if fullout is True (default).
    """
    _checkq(q)
    _checkm(m)
    l1, l1pot = get_lagrange_1(q, eps=eps)
    l2, _ = get_lagrange_2(q, eps=eps)
    l3, _ = get_lagrange_3(q, eps=eps)
    f = lambda x: rochepot_dl(x,0,0,q) - l1pot
    if m == 2:
        # Outer limit (beyond m2) of Roche lobe
        p1 = l1
        p2 = sco.brentq(f, 1+eps, l2)
        r = max(1-p1, p2-1)
    elif m == 1:
        # Outer limit (behind m1) of Roche lobe
        p1 = sco.brentq(f, l3, -eps)
        p2 = l1
        r = max(abs(p1), abs(p2))
    
    x = np.random.random(n)*2*r-r + (m-1)
    y = np.random.random(n)*2*r-r
    z = np.random.random(n)*2*r-r
    
    if pl is None:
        pl = l1pot
    
    ps = rochepot_dl(x, y, z, q)
    indi = np.where(ps > pl)[0]
    ef = len(indi)/n
    
    # Get volume of rotational body
    vol = (2*r)**3 * len(indi)/n
    dvol = (2*r)**3 * np.sqrt( (ef - ef**2)/n )
    if fullout:
        reff = (vol*3./4/np.pi)**(1./3)
        refferr = reff/(3*vol)*dvol
        return vol, dvol, reff, refferr
    return vol, dvol

def rochepot(x, y, z, m1, m2, a):
    r"""
    Roche potential (:math:`\Phi`, synchronous rotation)
    
    More massive component (:math:`m_1`) is at (x,y,z) = (0,0,0). Less massive
    component (:math:`m_2`) is at (a,0,0).
    
    Parameters
    ----------
    x, y, z : float or array
        Position [m]
    a : float
        Separation of masses [m]
    m1, m2 : float
        Masses of objects [kg] (m1 >= m2)
    
    Returns
    -------
    Roche potential : float or array
        At specified locations [m**2/s**2]
    """
    # Gravitational constant (m**3/(kg*s**2))
    G = 6.67430e-11
    M = m1 + m2
    return -(G*M)/(2*a) * rochepot_dl(x/a, y/a, z/a, m2/m1)
    
def coriolisAcceleration(omega,v):
    r"""
    Get Coriolis acceleration
    
    In the rotating frame, moving material is subject to the Coriolis
    force, determined by the circular frequency (vector) of the rotating frame
    and the velocity vector within it.
    
    .. math::
    
        a_C = -2 \vec{\omega} \times \vec{v}
    
    Parameters
    ----------
    omega : float or 3-element 1d array
        Circular frequency. If float is given, vector assumed to be omega*(0,0,1).
    v : array
        Either 3-element 1d array with velocity vector in corotating frame
        or array with shape (nv, 3) with nv velocity vectors (i.e., [::,0] gives all
        x components of velocity)
    
    Returns
    -------
    Coriolis acceleration : array
        Either 3-element array or [vn,3] array, depending in input shape of v (units
        determined by input).
    """
    if isinstance(omega, float):
        omvec = np.array([0,0,omega]) 
    else:
        omvec = omega
    c = -2*np.cross(omvec, v)
    return c
    
def roche_limit_fluid(q, R2):
    r"""
    Calculate the Roche limit for a fluid body on circular orbit
    
    An expression for the Roche limit, :math:`a_R`, was given by the eponymous
    Édouard Albert Roche in 1849 (see, e.g.,
    `Rappaport et al. 2013, ApJL 773 <https://ui.adsabs.harvard.edu/abs/2013ApJ...773L..15R/abstract>`_
    for a modern reference)
    
    .. math::
    
        a_R = 2.44 R_1 \sqrt[3]{\frac{\rho_1}{\rho_2}} = 2.44 R_2 q^{-1/3}
        
    where :math:`R_{1,2}` are the radii of the primary and secondary bodies, :math:`\rho_{1,2}`
    the respective densities, and :math:`q` the mass ratio; the second equality relies on spherical
    shapes.
    
    Parameters
    ----------
    q : float
        Mass ratio
    R2 : float
        Radius of secondary body (assuming spherical shape). Units are
        arbitrary and same as result.
    
    Returns
    -------
    aR : float
        The Roche distance in the units of R2.
    """
    _checkq(q)
    return 2.44*R2*q**(-1/3)
    
    



import numpy as np


def convertDampingConstant(gamma, wavelength):
    r"""
    Convert damping constant into line-width in wavelength units.

    The inverse lifetime, :math:`1/\tau`, is related to the damping constant,
    :math:`\Gamma`, and the Einstein coefficients, A, according to  

    .. math::

        \frac{1}{\tau_{lu}} = \Gamma_{lu} = \sum_{j<u} A_{lj} + \sum_{j<l} A_{uj}\,,

    where u and l denote the upper and lower state of the transition (see, e.g.,
    "Der neue Kosmos", Unsoeld und Baschek, 7th edition).
    The damping coefficient, `gamma`, determines the FWHM in units of
    circular frequency.
    The natural line width (FWHM of the Lorentzian profile) in wavelength
    units is then given by

    .. math::

        \frac{\lambda^2}{2\pi c}\Gamma\,.

    Parameters
    ----------
    gamma : float 
        The damping constant [1/s]. The damping constant may be determined
        by summing all relevant Einstein coefficients as shown above.
    wavelength : float
        Wavelength of the transition in A.

    Returns
    -------
    Natural line-width : float
        The Full Width at Half Maximum of the Lorentzian [cm]. 
    """
    # The factor of 1e-8 transforms Angstrom into cm
    return gamma * (wavelength*1e-8)**2 / (2.0*np.pi*29979245800.)



from PyAstronomy import funcFit as fuf
import numpy as np


class VoigtAstroP(fuf.OneDFit):
    r"""
    Astronomical parameterization of Voigt profile.

    This class provides a convenient parameterization of the Voigt
    profile, frequently used in astronomy. In particular,
    the line is parameterized in terms of wavelength,
    Doppler parameter, damping width, and oscillator strength.

    If the velocity components of the constituents follow
    independent Gaussian distributions with standard deviation
    :math:`\sigma` in all three dimensions of space

    .. math::

        v_{x,y,z} \sim N(0,\sigma^2) \; ,

    the mode of the distribution of absolute velocity,
    :math:`v_a = \sqrt{v_x^2 + v_y^2 + v_z^2}`
    is given by :math:`v_{mode} = \sqrt{2}\sigma`. This is
    the velocity referred to as the Doppler parameter, b.
    Consequently, the velocity dispersion in any individual dimension,
    i.e., the standard deviation of the Gaussian velocity distribution,
    is related to the Doppler parameter by :math:`\sigma = b/\sqrt{2}`.
    For zero damping width (gamma),
    the resulting model line is, therefore, a Gaussian with a standard
    deviation of :math:`\sigma = b/\sqrt{2}` in velocity units.

    Instrumental resolution can be applied via the parameter R.
    The instrumental profile is assumed to be a Gaussian with
    FWHM of w0/R. The additional broadening is implemented by
    using an internal, effective Doppler parameter. The case
    R=0 corresponds to infinite instrumental resolution (i.e.,
    no additional broadening).

    *Fit parameters*:

    ======= ================================================ =====

    w0      Wavelength of the transition                     A
    b       Doppler parameter (corresponds                   km/s
            to sqrt(2) times the velocity dispersion).
    gamma   Damping width (full width at half maximum of
            the Lorentzian)                                  cm
    f       Oscillator strength (unitless)                   --
    R       Instrumental resolution                          --
    ======= ================================================ =====

    """

    def __init__(self):
        fuf.OneDFit.__init__(
            self, ["w0", "b", "gamma", "f", "lin", "off", "R"], rootName="VoigtAstroP")
        self._profile = fuf.Voigt1d()
        # Set default parameter, which may not be changed
        self._profile["ad"] = 1. / np.sqrt(2.0)
        # Set to zero, because evaluation is done in velocity space
        self._profile["mu"] = 0.0
        # Conversion factor between Gaussian FWHM and STD
        self._fwhmstd = 2*np.sqrt(2*np.log(2.))

    def bl(self):
        r"""
        Doppler width in cm
        """
        bl = self["w0"] / 1e8 * (self["b"] * 100000.0) / 29979245800.0
        if self["R"] > 0:
            # Account for instrumental resolution
            bl = np.sqrt(
                bl**2 + 2*(self["w0"]/(1e8*self["R"]*self._fwhmstd))**2)
        return bl

    def FWHM(self):
        r"""
        Estimate FWHM

        Applies same approximation is Voigt1d

        Returns
        -------
        FWHM : float
            FWHM of line profile in wavelength units [A]
        """
        fwhm = self._profile.FWHM()
        c = 1e8 * self.bl()
        return fwhm * c

    def evaluate(self, x):
        r"""
        Evaluate the absorption-line profile.

        Parameters
        ----------
        x : array of floats
            Contains the wavelength values in Angstrom.

        Returns
        -------
        Model : array of floats
            Return the cross-section in cm^2.
        """
        # Wavelength in cm
        w0cm = self["w0"] / 1e8
        # Doppler width in cm
        bl = self.bl()
        # The following formulation is, e.g., according to D.F. Gray "Stellar photospheres"
        # (third edition, Eq. 11.46, p. 257). For the relation between H(u, a) and the Voigt
        # profile see Pagnini and Saxena "A note on the Voigt profile" (Eq. 6,
        # https://arxiv.org/abs/0805.2274); the relation quoted by Gray is probably not entirely
        # accurate. Note that the Voigt profile is here evaluated at u = dw/bl.
        #
        # The constant equals (pi e^2)/(m_e c^2) (pi times the classical electron radius)
        self._profile["A"] = 8.85282064473e-13 * self["f"] * w0cm**2 / bl
        # A factor of 2.0 because `al` defines the half FWHM in Voigt profile (division by bl
        # accounts for evaluation at dw/bl).
        self._profile["al"] = (self["gamma"] / bl) / 2.0
        self._profile["lin"] = self["lin"]
        self._profile["off"] = self["off"]
        # Convert the given wavelength axis into velocity units
        u = (x / 1e8 - w0cm) / bl
        return self._profile.evaluate(u)



import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import _ic
import os


def read1dFitsSpec(fn, hdu=0, fullout=False, CRPIX1=None, keymap={}):
    """
    Read a simple 1d spectrum from fits file.

    Reads a 1d-spectrum from a file and constructs the
    associated wavelength axis. To this end, the expression:
    wvl = ((np.arange(N) + 1.0) - CRPIX1) * CDELT1 + CRVAL1
    will be evaluated, where N is the number of bins and
    CRPIX1, CDELT1, and CRVAL1 are header keywords.

    Parameters
    ----------
    fn : string
        Filename
    hdu : int, optional
        The number of the HDU to be used. The default
        is 0, i.e., the primary HDU.
    fullout : boolean, optional
        If True, the header keywords used to construct
        the wavelength axis will be returned. The default
        is False.
    CRPIX1 : int, optional
        Can be used to circumvent missing CRPIX1 entry.
    keymap : dict, optional
        Can be used to map header keywords

    Returns
    -------
    wvl : array
        The wavelength array.
    flx : array
        The flux array.

    Examples
    --------

    ::

      from PyAstronomy import pyasl
      wvl, flx = pyasl.read1dFitsSpec("mySpectrum.fits")
    """
    if (not _ic.check["pyfits"]) and (not _ic.check["astropy.io.fits"]):
        raise(PE.PyARequiredImport("Could neither import module 'pyfits' and 'astropy.io.fits'.",
                                   where="read1dFitsSpec",
                                   solution="Install pyfits: http://www.stsci.edu/institute/software_hardware/pyfits"))

    if not os.path.isfile(fn):
        raise(PE.PyAFileError(fn, "ne",
                              where="read1dFitsSpec",
                              solution="Check file name."))

    if _ic.check["pyfits"]:
        import pyfits
    else:
        import astropy.io.fits as pyfits

    hl = pyfits.open(fn)

    naxis = hl[hdu].header["NAXIS"]
    if hl[hdu].header["NAXIS"] != 1:
        if hl[hdu].header["NAXIS"] == 2 and hl[hdu].header["NAXIS2"] == 1:
            pass
        else:
            raise(PE.PyAValError("There is more than one axis (NAXIS = " + str(naxis) + ", actual value is " + str(hl[hdu].header["NAXIS"]) + ").",
                                 where="read1dFitsSpec",
                                 solution="Check file and HDU."))

    hkeys = {"CRVAL1": None, "CRPIX1": CRPIX1, "CDELT1": None}
    for k in hkeys.keys():
        if k not in keymap:
            keymap[k] = k

    for k in hkeys.keys():
        if hkeys[k] is not None:
            continue
        if not keymap[k] in hl[hdu].header:
            raise(PE.PyAValError("Header does not contain required keyword '" + str(keymap[k]) + "'",
                                 where="read1dFitsSpec",
                                 solution="Check file and HDU."))
        hkeys[k] = hl[hdu].header[keymap[k]]

    N = int(hl[hdu].header["NAXIS1"])

    # Construct wvl axis
    wvl = ((np.arange(N) + 1.0) - hkeys["CRPIX1"]
           ) * hkeys["CDELT1"] + hkeys["CRVAL1"]
    # Get flux
    if naxis == 1:
        flx = np.array(hl[hdu].data)
    elif naxis == 2:
        flx = np.array(hl[hdu].data[0])

    if fullout:
        # Full output
        return wvl, flx, hkeys

    return wvl, flx



# -*- coding: utf-8 -*-
from __future__ import print_function
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyaC import ImportCheck
import sys

# Check Python version
_majV, _minV = sys.version_info[0:2]
if (_minV < 7) and (_majV == 2):
    PE.warn("funcFit needs 2.7.x or greater. See documentation (Prerequisites) for explanation.")

ic = ImportCheck(["numpy", "scipy", "pymc", "matplotlib",
                  "matplotlib.pylab", "pyfits", "emcee", "progressbar"])

# Get out if numpy not present
if not ic.check["numpy"]:
    raise(PE.PyARequiredImport("Numpy cannot be imported.", solution="Install numpy (see http://numpy.scipy.org/, you probably should also install SciPy).",
                               addInfo="The numpy package provides array support for Python and is indispensable in many scientific applications."))

# Check whether fitting modules can be imported
_scoImport = ic.check["scipy"]
_pymcImport = ic.check["pymc"]
_mplImport = ic.check["matplotlib"]

from PyAstronomy.funcFit.utils import *
from .modelRebin import turnIntoRebin, _ModelRebinDocu
from .onedfit import *
from .gauss1d import *
from .gauss2d import *
from .params import *
from .sinexp1d import *
from .polyModel import *
from .cauchyLorentz1d import *
# from .anneal import *
from .syncFit import SyncFitContainer
from .coordinateGrid import *
from .circle2d import *
from .sech import *
from .extFitter import NelderMead
from .fufDS import FufDS

if ic.check["scipy"]:
    from .voigt1d import *
else:
    PE.warn(PE.PyARequiredImport("No Voigt profile available, because scipy cannot be imported",
                                 solution="Install scipy."))


def __question(tf):
    if tf:
        return "yes"
    return "no"


def status():
    print("Status of funcFit:")
    print("--------------------------")
    print("Is scipy.optimize available? ", __question(_scoImport))
    if not _scoImport:
        print("  The minimization algorithms from SciPy cannot be used.")
        print("  Install SciPy to make them available (see http://www.scipy.org/).")
    print("Is pymc available? ", __question(_pymcImport))
    if not _pymcImport:
        print("  No Markov-Chain Monte-Carlo sampling via pymc will be possible.")
        print("  Install pymc to enable it (see http://code.google.com/p/pymc/).")
    else:
        import pymc
        print("  pymc is available in version: ", pymc.__version__)
        try:
            majorversion = int(pymc.__version__.split('.')[0])
            if majorversion > 2:
                print("  NOTE: Sampling is only supported with pymc 2.x (not 3.x)!")
        except:
            print("  Could not parse pymc version string...")
    print("Is emcee available? ", __question(ic.check["emcee"]))
    if not ic.check["emcee"]:
        print("  No Markov-Chain Monte-Carlo sampling via emcee will be possible.")
        print("  Install emcee to enable it (see http://dan.iel.fm/emcee/current/).")


__all__ = ["cauchyLorentz1d", "gauss1d", "modelRebin", "onedfit", "params", "voigt1d", "sinexp1d", "polyModel",
           "BoltzmannSAS", "PyAAnneal"]



from __future__ import print_function, division
import unittest


class TestStringlengthSanity(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testsanity_slExample(self):
        """Checking stringlength example"""
        import numpy as np
        import matplotlib.pylab as plt
        from PyAstronomy import pyTiming as pyt

        period = 1.75

        x = np.linspace(0, 10, 100)
        y = 20 * np.sin(2 * np.pi * x / period)
        y += np.random.normal(0, 1, len(x))

        # Trial periods to be tested (200 trial periods between 0.5 and 4.5;
        # same units as x-axis)
        tps = (0.5, 4.5, 200)

        # Calculate string length
        p, sl = pyt.stringlength_dat(x, y, tps)

        # Show the string length. An alias at the double period
        # (half frequency) is obvious.

    #         plt.plot(p, sl, 'b.-')
    #         plt.ylabel("String length")
    #         plt.xlabel("Trial period")
    #         plt.show()

    def testsanity_stringlength(self):
        """Checking sanity of stringlength"""

        import numpy as np
        from PyAstronomy import pyTiming as pyt

        x = np.array([0.0, 0.5, 0.999999])
        y = np.array([0.0, 0.5, 1.0])

        # Calculate string length
        p, sl = pyt.stringlength_dat(x, y, tps=np.array([1]), norm="no")
        self.assertAlmostEqual(
            sl[0],
            np.sqrt(2.0) + 1.0,
            delta=1e-6,
            msg="closed stringlength does not match",
        )

        p, sl = pyt.stringlength_dat(x, y, tps=np.array([1]), norm="no", closed=False)
        self.assertAlmostEqual(
            sl[0],
            np.sqrt(2.0),
            delta=1e-6,
            msg="stringlength (not closed) does not match",
        )



from __future__ import print_function
from . import pyaErrors as PE
import difflib
import six
 

def fuzzyMatch(inkey, wordList, caseSensitive=True, n=3, cutoff=0.6, raises=False):
  """
    Find exact and approximate matches for input in word list.
    
    Uses `get_close_matches` from Python's difflib to search
    for the best matches between the input keyword and a list
    of words.
    
    Parameters
    ----------
    inkey : string
        The input keyword.
    wordList : list of strings
        List of words with possible matches for `inkey`.
    caseSensitive : boolean, optional
        If True (default), the search will be case
        sensitive.
    n : int, optional
        Number of potential matches returned by
        `get_close_matches`.
    cutoff : float, optional
        Number between 0 and 1 indicating the degree
        of similarity between `inkey` and the entries from
        `wordList`. The lower the number, the more dissimilar
        the potential matches may be. (`cutoff` parameter
        from `get_close_matches`).
    raises : boolean, optional
        If True, a `PyAValError` giving a summary of the
        failure will be raised if no exact match is found.
        The default is false.
    
    Returns
    -------
    Matches : dictionary
        If found, contains the exact match (key "em") found
        in the list (in lower case if `caseSensitive` is True)
        and a list of close matches (key "cm"), which the user
        may have meant to specify.  If an exact match is found,
        also its index in `wordList` is provided.
  """
  if not isinstance(wordList[0], six.string_types):
    raise(PE.PyAValError("Worlist does not seem to be a list of strings. First item has type: " + str(type(wordList[0])), \
                         where="fuzzyMatch", \
                         solution="Convert into string."))
  if not caseSensitive:
    key = inkey.lower()
    wordList = [x.lower() for x in wordList]
  else:
    key = inkey
  result = {"em":None, "cm":None, "index":None}
  result["cm"] = difflib.get_close_matches(key, wordList, n, cutoff)
  if len(result["cm"]) == 0:
    if raises:
      raise(PE.PyAValError("No match for '" + str(inkey) + "'", \
            solution="Check the input keyword and/or the data."))
  if key == result["cm"][0]:
    # An exact match between the input list
    result["em"] = key
    result["index"] = wordList.index(key)
  else:
    # There is no exact, but a approximate matches
    if raises:
      raise(PE.PyAValError("No exact match found for key '" + str(inkey) + "'\n" + \
                           "Did you mean any of these: " + ', '.join(result["cm"]) + "?", \
              solution=["Check the input keyword and/or the data."]))
  return result


from __future__ import division
import numpy as np
from PyAstronomy.pyaC import pyaErrors as PE

def zerocross1d(x, y, getIndices=False):
  """
    Find the zero crossing points in 1d data.
    
    Find the zero crossing events in a discrete data set.
    Linear interpolation is used to determine the actual
    locations of the zero crossing between two data points
    showing a change in sign. Data point which are zero
    are counted in as zero crossings if a sign change occurs
    across them. Note that the first and last data point will
    not be considered whether or not they are zero. 
    
    Parameters
    ----------
    x, y : arrays
        Ordinate and abscissa data values.
    getIndices : boolean, optional
        If True, also the indicies of the points preceding
        the zero crossing event will be returned. Defeualt is
        False.
    
    Returns
    -------
    xvals : array
        The locations of the zero crossing events determined
        by linear interpolation on the data.
    indices : array, optional
        The indices of the points preceding the zero crossing
        events. Only returned if `getIndices` is set True.
  """
  
  # Check sorting of x-values
  if np.any((x[1:] - x[0:-1]) <= 0.0):
    raise(PE.PyAValError("The x-values must be sorted in ascending order!", \
                         where="zerocross1d", \
                         solution="Sort the data prior to calling zerocross1d."))
  
  # Indices of points *before* zero-crossing
  indi = np.where(y[1:]*y[0:-1] < 0.0)[0]
  
  # Find the zero crossing by linear interpolation
  dx = x[indi+1] - x[indi]
  dy = y[indi+1] - y[indi]
  zc = -y[indi] * (dx/dy) + x[indi]
  
  # What about the points, which are actually zero
  zi = np.where(y == 0.0)[0]
  # Do nothing about the first and last point should they
  # be zero
  zi = zi[np.where((zi > 0) & (zi < x.size-1))]
  # Select those point, where zero is crossed (sign change
  # across the point)
  zi = zi[np.where(y[zi-1]*y[zi+1] < 0.0)]
  
  # Concatenate indices
  zzindi = np.concatenate((indi, zi)) 
  # Concatenate zc and locations corresponding to zi
  zz = np.concatenate((zc, x[zi]))
  
  # Sort by x-value
  sind = np.argsort(zz)
  zz, zzindi = zz[sind], zzindi[sind]
  
  if not getIndices:
    return zz
  else:
    return zz, zzindi
  


from __future__ import absolute_import
from .pyaErrTemplate import PyaErrTemplate

class PyAValError(PyaErrTemplate):
  
  def __init__(self, what, **keys):
    """
      When to be raised?
      
      Whenever an unexpected value is encountered.
    """
    PyaErrTemplate.__init__(self, what, "PyA Value Error", **keys)


class PyANameClash(PyaErrTemplate):

  def __init__(self, what, **keys):
    """
      When to be raised?
      
      Whenever an unexpected doubling of names (e.g., key values in a dictionary) is encountered. 
    """
    PyaErrTemplate.__init__(self, what, "PyA Name Clash", **keys)


class PyAFloatingPointError(PyaErrTemplate):

  def __init__(self, what, **keys):
    """
      When to be raised?
      
      Whenever a floating point error is caught.
    """
    PyaErrTemplate.__init__(self, what, "PyA Floating Point Error", **keys)